import itertools
import os
import pytest
import unittest
from copy import deepcopy
from itertools import product
from parameterized import parameterized

import h5py
import numpy as np
import bilby
from bilby.gw.likelihood import BilbyROQParamsRangeError


class TestBasicGWTransient(unittest.TestCase):
    def setUp(self):
        np.random.seed(500)
        self.parameters = dict(
            mass_1=31.0,
            mass_2=29.0,
            a_1=0.4,
            a_2=0.3,
            tilt_1=0.0,
            tilt_2=0.0,
            phi_12=1.7,
            phi_jl=0.3,
            luminosity_distance=4000.0,
            theta_jn=0.4,
            psi=2.659,
            phase=1.3,
            geocent_time=1126259642.413,
            ra=1.375,
            dec=-1.2108,
        )
        self.interferometers = bilby.gw.detector.InterferometerList(["H1"])
        self.interferometers.set_strain_data_from_power_spectral_densities(
            sampling_frequency=2048, duration=4
        )
        self.waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
            duration=4,
            sampling_frequency=2048,
            frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
        )

        self.likelihood = bilby.gw.likelihood.BasicGravitationalWaveTransient(
            interferometers=self.interferometers,
            waveform_generator=self.waveform_generator,
        )
        self.likelihood.parameters = self.parameters.copy()

    def tearDown(self):
        del self.parameters
        del self.interferometers
        del self.waveform_generator
        del self.likelihood

    def test_noise_log_likelihood(self):
        """Test noise log likelihood matches precomputed value"""
        self.likelihood.noise_log_likelihood()
        self.assertAlmostEqual(
            -4037.0994372143414, self.likelihood.noise_log_likelihood(), 3
        )

    def test_log_likelihood(self):
        """Test log likelihood matches precomputed value"""
        self.likelihood.log_likelihood()
        self.assertAlmostEqual(self.likelihood.log_likelihood(), -4054.047229508672, 3)

    def test_log_likelihood_ratio(self):
        """Test log likelihood ratio returns the correct value"""
        self.assertAlmostEqual(
            self.likelihood.log_likelihood() - self.likelihood.noise_log_likelihood(),
            self.likelihood.log_likelihood_ratio(),
            3,
        )

    def test_likelihood_zero_when_waveform_is_none(self):
        """Test log likelihood returns np.nan_to_num(-np.inf) when the
        waveform is None"""
        self.likelihood.waveform_generator.frequency_domain_strain = lambda x: None
        self.assertEqual(self.likelihood.log_likelihood_ratio(), np.nan_to_num(-np.inf))

    def test_repr(self):
        expected = "BasicGravitationalWaveTransient(interferometers={},\n\twaveform_generator={})".format(
            self.interferometers, self.waveform_generator
        )
        self.assertEqual(expected, repr(self.likelihood))


class TestGWTransient(unittest.TestCase):
    def setUp(self):
        np.random.seed(500)
        self.duration = 4
        self.sampling_frequency = 2048
        self.parameters = dict(
            mass_1=31.0,
            mass_2=29.0,
            a_1=0.4,
            a_2=0.3,
            tilt_1=0.0,
            tilt_2=0.0,
            phi_12=1.7,
            phi_jl=0.3,
            luminosity_distance=4000.0,
            theta_jn=0.4,
            psi=2.659,
            phase=1.3,
            geocent_time=1126259642.413,
            ra=1.375,
            dec=-1.2108,
        )
        self.interferometers = bilby.gw.detector.InterferometerList(["H1"])
        self.interferometers.set_strain_data_from_power_spectral_densities(
            sampling_frequency=self.sampling_frequency, duration=self.duration
        )
        self.waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
            duration=self.duration,
            sampling_frequency=self.sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
        )

        self.prior = bilby.gw.prior.BBHPriorDict()
        self.prior["geocent_time"] = bilby.prior.Uniform(
            minimum=self.parameters["geocent_time"] - self.duration / 2,
            maximum=self.parameters["geocent_time"] + self.duration / 2,
        )

        self.likelihood = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=self.interferometers,
            waveform_generator=self.waveform_generator,
            priors=self.prior.copy(),
        )
        self.likelihood.parameters = self.parameters.copy()

    def tearDown(self):
        del self.parameters
        del self.interferometers
        del self.waveform_generator
        del self.prior
        del self.likelihood

    def test_noise_log_likelihood(self):
        """Test noise log likelihood matches precomputed value"""
        self.likelihood.noise_log_likelihood()
        self.assertAlmostEqual(
            -4037.0994372143414, self.likelihood.noise_log_likelihood(), 3
        )

    def test_log_likelihood(self):
        """Test log likelihood matches precomputed value"""
        self.likelihood.log_likelihood()
        self.assertAlmostEqual(self.likelihood.log_likelihood(),
                               -4054.047229508673, 3)

    def test_log_likelihood_ratio(self):
        """Test log likelihood ratio returns the correct value"""
        self.assertAlmostEqual(
            self.likelihood.log_likelihood() - self.likelihood.noise_log_likelihood(),
            self.likelihood.log_likelihood_ratio(),
            3,
        )

    def test_likelihood_zero_when_waveform_is_none(self):
        """Test log likelihood returns np.nan_to_num(-np.inf) when the
        waveform is None"""
        self.likelihood.waveform_generator.frequency_domain_strain = lambda x: None
        self.assertEqual(self.likelihood.log_likelihood_ratio(), np.nan_to_num(-np.inf))

    def test_repr(self):
        expected = (
            "GravitationalWaveTransient(interferometers={},\n\twaveform_generator={},\n\t"
            "time_marginalization={}, distance_marginalization={}, phase_marginalization={}, "
            "calibration_marginalization={}, priors={})".format(
                self.interferometers,
                self.waveform_generator,
                False,
                False,
                False,
                False,
                self.prior,
            )
        )
        self.assertEqual(expected, repr(self.likelihood))

    def test_interferometers_setting_list(self):
        ifos = [
            bilby.gw.detector.get_empty_interferometer(name=name)
            for name in ["H1", "L1"]
        ]
        self.likelihood.interferometers = ifos
        self.assertListEqual(
            bilby.gw.detector.InterferometerList(ifos), self.likelihood.interferometers
        )
        self.assertTrue(
            type(self.likelihood.interferometers)
            == bilby.gw.detector.InterferometerList
        )

    def test_interferometers_setting_interferometer_list(self):
        ifos = bilby.gw.detector.InterferometerList(
            [
                bilby.gw.detector.get_empty_interferometer(name=name)
                for name in ["H1", "L1"]
            ]
        )
        self.likelihood.interferometers = ifos
        self.assertListEqual(
            bilby.gw.detector.InterferometerList(ifos), self.likelihood.interferometers
        )
        self.assertTrue(
            type(self.likelihood.interferometers)
            == bilby.gw.detector.InterferometerList
        )

    def test_meta_data(self):
        expected = dict(
            interferometers=self.interferometers.meta_data,
            time_marginalization=False,
            phase_marginalization=False,
            distance_marginalization=False,
            calibration_marginalization=False,
            waveform_generator_class=self.waveform_generator.__class__,
            waveform_arguments=self.waveform_generator.waveform_arguments,
            frequency_domain_source_model=self.waveform_generator.frequency_domain_source_model,
            parameter_conversion=self.waveform_generator.parameter_conversion,
            sampling_frequency=self.waveform_generator.sampling_frequency,
            duration=self.waveform_generator.duration,
            start_time=self.waveform_generator.start_time,
            time_reference="geocent",
            reference_frame="sky",
            lal_version=self.likelihood.lal_version,
            lalsimulation_version=self.likelihood.lalsimulation_version,
        )
        self.assertDictEqual(expected, self.likelihood.meta_data)

    def test_reference_frame_agrees_with_default(self):
        new_likelihood = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=self.interferometers,
            waveform_generator=self.waveform_generator,
            priors=self.prior.copy(),
            reference_frame="H1L1"
        )
        parameters = self.parameters.copy()
        del parameters["ra"], parameters["dec"]
        parameters["zenith"] = 1.0
        parameters["azimuth"] = 1.0
        parameters["ra"], parameters["dec"] = bilby.gw.utils.zenith_azimuth_to_ra_dec(
            zenith=parameters["zenith"],
            azimuth=parameters["azimuth"],
            geocent_time=parameters["geocent_time"],
            ifos=bilby.gw.detector.InterferometerList(["H1", "L1"])
        )
        new_likelihood.parameters.update(parameters)
        self.likelihood.parameters.update(parameters)
        self.assertEqual(
            new_likelihood.log_likelihood_ratio(),
            self.likelihood.log_likelihood_ratio()
        )

    def test_time_reference_agrees_with_default(self):
        new_likelihood = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=self.interferometers,
            waveform_generator=self.waveform_generator,
            priors=self.prior.copy(),
            time_reference="H1"
        )
        ifo = bilby.gw.detector.get_empty_interferometer("H1")
        time_delay = ifo.time_delay_from_geocenter(
            ra=self.parameters["ra"],
            dec=self.parameters["dec"],
            time=self.parameters["geocent_time"]
        )
        parameters = self.parameters.copy()
        parameters["H1_time"] = parameters["geocent_time"] + time_delay
        new_likelihood.parameters.update(parameters)
        self.likelihood.parameters.update(parameters)
        self.assertEqual(
            new_likelihood.log_likelihood_ratio(),
            self.likelihood.log_likelihood_ratio()
        )


class TestMarginalizedLikelihood(unittest.TestCase):
    def setUp(self):
        np.random.seed(500)
        self.duration = 4
        self.sampling_frequency = 2048
        self.parameters = dict(
            mass_1=31.0,
            mass_2=29.0,
            a_1=0.4,
            a_2=0.3,
            tilt_1=0.0,
            tilt_2=0.0,
            phi_12=1.7,
            phi_jl=0.3,
            luminosity_distance=4000.0,
            theta_jn=0.4,
            psi=2.659,
            phase=1.3,
            geocent_time=1126259642.413,
            ra=1.375,
            dec=-1.2108,
        )

        self.interferometers = bilby.gw.detector.InterferometerList(["H1"])
        self.interferometers.set_strain_data_from_power_spectral_densities(
            sampling_frequency=self.sampling_frequency,
            duration=self.duration,
            start_time=self.parameters["geocent_time"] - self.duration / 2,
        )

        self.waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
            duration=self.duration,
            sampling_frequency=self.sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
        )

        self.prior = bilby.gw.prior.BBHPriorDict()
        self.prior["geocent_time"] = bilby.prior.Uniform(
            minimum=self.parameters["geocent_time"] - self.duration / 2,
            maximum=self.parameters["geocent_time"] + self.duration / 2,
        )

    def test_cannot_instantiate_marginalised_likelihood_without_prior(self):
        self.assertRaises(
            ValueError,
            lambda: bilby.gw.likelihood.GravitationalWaveTransient(
                interferometers=self.interferometers,
                waveform_generator=self.waveform_generator,
                phase_marginalization=True,
            ),
        )

    def test_generating_default_time_prior(self):
        temp = self.prior.pop("geocent_time")
        new_prior = self.prior.copy()
        like = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=self.interferometers,
            waveform_generator=self.waveform_generator,
            priors=new_prior,
            time_marginalization=True,
        )
        same = all(
            [
                temp.minimum == like.priors["geocent_time"].minimum,
                temp.maximum == like.priors["geocent_time"].maximum,
                new_prior["geocent_time"] == temp.minimum,
            ]
        )
        self.assertTrue(same)
        self.prior["geocent_time"] = temp

    def test_generating_default_phase_prior(self):
        temp = self.prior.pop("phase")
        new_prior = self.prior.copy()
        like = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=self.interferometers,
            waveform_generator=self.waveform_generator,
            priors=new_prior,
            phase_marginalization=True,
        )
        same = all(
            [
                temp.minimum == like.priors["phase"].minimum,
                temp.maximum == like.priors["phase"].maximum,
                new_prior["phase"] == float(0),
            ]
        )
        self.assertTrue(same)
        self.prior["phase"] = temp

    def test_run_sampler_flags_if_marginalized_phase_is_sampled(self):
        like = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=self.interferometers,
            waveform_generator=self.waveform_generator,
            priors=self.prior,
            phase_marginalization=True,
        )
        new_prior = self.prior.copy()
        new_prior["phase"] = bilby.prior.Uniform(minimum=0, maximum=2 * np.pi)
        for key, param in dict(
            mass_1=31.0,
            mass_2=29.0,
            a_1=0.4,
            a_2=0.3,
            tilt_1=0.0,
            tilt_2=0.0,
            phi_12=1.7,
            phi_jl=0.3,
            theta_jn=0.4,
            psi=2.659,
            ra=1.375,
            dec=-1.2108,
        ).items():
            new_prior[key] = param
        with self.assertRaises(bilby.core.sampler.SamplingMarginalisedParameterError):
            bilby.run_sampler(like, new_prior)

    def test_run_sampler_flags_if_marginalized_time_is_sampled(self):
        like = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=self.interferometers,
            waveform_generator=self.waveform_generator,
            priors=self.prior,
            time_marginalization=True,
        )
        new_prior = self.prior.copy()
        new_prior["geocent_time"] = bilby.prior.Uniform(minimum=0, maximum=1)
        for key, param in dict(
            mass_1=31.0,
            mass_2=29.0,
            a_1=0.4,
            a_2=0.3,
            tilt_1=0.0,
            tilt_2=0.0,
            phi_12=1.7,
            phi_jl=0.3,
            theta_jn=0.4,
            psi=2.659,
            ra=1.375,
            dec=-1.2108,
        ).items():
            new_prior[key] = param
        with self.assertRaises(bilby.core.sampler.SamplingMarginalisedParameterError):
            bilby.run_sampler(like, new_prior)


class TestMarginalizations(unittest.TestCase):
    """
    Test all marginalised likelihoods matches brute force version.

    For time, this is strongly dependent on the specific time grid used.
    The `time_jitter` parameter makes this a weaker dependence during sampling.
    """
    _parameters = product(
        ["regular", "roq"],
        ["luminosity_distance", "geocent_time", "phase"],
        [True, False],
        [True, False],
        [True, False],
    )

    lookup_phase = "distance_lookup_phase.npz"
    lookup_no_phase = "distance_lookup_no_phase.npz"
    path_to_roq_weights = "weights.npz"

    def setUp(self):
        np.random.seed(500)
        self.duration = 4
        self.sampling_frequency = 2048
        self.parameters = dict(
            mass_1=31.0,
            mass_2=29.0,
            a_1=0.4,
            a_2=0.3,
            tilt_1=0.0,
            tilt_2=0.0,
            phi_12=1.7,
            phi_jl=0.3,
            luminosity_distance=4000.0,
            theta_jn=0.4,
            psi=2.659,
            phase=1.3,
            geocent_time=1126259642.413,
            ra=1.375,
            dec=-1.2108,
            time_jitter=0,
        )

        self.interferometers = bilby.gw.detector.InterferometerList(["H1"])
        self.interferometers.set_strain_data_from_power_spectral_densities(
            sampling_frequency=self.sampling_frequency,
            duration=self.duration,
            start_time=1126259640,
        )

        self.waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
            duration=self.duration,
            sampling_frequency=self.sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
            start_time=1126259640,
            waveform_arguments=dict(
                reference_frequency=20.0,
                minimum_frequency=20.0,
                approximant="IMRPhenomPv2",
            )
        )
        self.interferometers.inject_signal(
            parameters=self.parameters, waveform_generator=self.waveform_generator
        )

        self.priors = bilby.gw.prior.BBHPriorDict()
        self.priors["geocent_time"] = bilby.prior.Uniform(
            minimum=self.parameters["geocent_time"] - 0.1,
            maximum=self.parameters["geocent_time"] + 0.1
        )

        trial_roq_paths = [
            "/roq_basis",
            os.path.join(os.path.expanduser("~"), "ROQ_data/IMRPhenomPv2/4s"),
            "/home/cbc/ROQ_data/IMRPhenomPv2/4s",
        ]
        roq_dir = None
        for path in trial_roq_paths:
            if os.path.isdir(path):
                roq_dir = path
                break
        if roq_dir is None:
            raise Exception("Unable to load ROQ basis: cannot proceed with tests")

        self.roq_waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
            duration=self.duration,
            sampling_frequency=self.sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.binary_black_hole_roq,
            start_time=1126259640,
            waveform_arguments=dict(
                reference_frequency=20.0,
                minimum_frequency=20.0,
                approximant="IMRPhenomPv2",
                frequency_nodes_linear=np.load(f"{roq_dir}/fnodes_linear.npy"),
                frequency_nodes_quadratic=np.load(f"{roq_dir}/fnodes_quadratic.npy"),
            )
        )
        self.roq_linear_matrix_file = f"{roq_dir}/B_linear.npy"
        self.roq_quadratic_matrix_file = f"{roq_dir}/B_quadratic.npy"

    def tearDown(self):
        del self.duration
        del self.sampling_frequency
        del self.parameters
        del self.interferometers
        del self.waveform_generator
        del self.roq_waveform_generator
        del self.priors

    @classmethod
    def tearDownClass(cls):
        """remove lookup tables so that they are not used accidentally in subsequent tests"""
        for filename in [cls.lookup_phase, cls.lookup_no_phase, cls.path_to_roq_weights]:
            if os.path.exists(filename):
                os.remove(filename)

    def likelihood_kwargs(self, kind, time_marginalization, phase_marginalization, distance_marginalization, priors):
        if priors is None:
            priors = self.priors.copy()
        if distance_marginalization and phase_marginalization:
            lookup = TestMarginalizations.lookup_phase
        elif distance_marginalization:
            lookup = TestMarginalizations.lookup_no_phase
        else:
            lookup = None
        kwargs = dict(
            interferometers=self.interferometers,
            waveform_generator=self.waveform_generator,
            distance_marginalization=distance_marginalization,
            phase_marginalization=phase_marginalization,
            time_marginalization=time_marginalization,
            distance_marginalization_lookup_table=lookup,
            priors=priors,
        )
        if kind == "roq":
            kwargs.update(dict(
                linear_matrix=self.roq_linear_matrix_file,
                quadratic_matrix=self.roq_quadratic_matrix_file,
                waveform_generator=self.roq_waveform_generator,
            ))
            if os.path.exists(self.__class__.path_to_roq_weights):
                kwargs["weights"] = self.__class__.path_to_roq_weights
        return kwargs

    def get_likelihood(
        self,
        kind,
        time_marginalization=False,
        phase_marginalization=False,
        distance_marginalization=False,
        priors=None
    ):
        kwargs = self.likelihood_kwargs(
            kind, time_marginalization, phase_marginalization, distance_marginalization, priors
        )
        if kind == "regular":
            cls_ = bilby.gw.likelihood.GravitationalWaveTransient
        elif kind == "roq":
            cls_ = bilby.gw.likelihood.ROQGravitationalWaveTransient
        else:
            raise ValueError(f"kind {kind} not understood")
        like = cls_(**kwargs)
        if kind == "roq" and not os.path.exists(self.__class__.path_to_roq_weights):
            like.save_weights(self.__class__.path_to_roq_weights)
        like.parameters = self.parameters.copy()
        if time_marginalization:
            like.parameters["geocent_time"] = self.interferometers.start_time
        return like

    def _template(self, marginalized, non_marginalized, key, prior=None, values=None):
        if prior is None:
            prior = self.priors[key]
        if values is None:
            values = np.linspace(prior.minimum, prior.maximum, 1000)
        prior_values = prior.prob(values)
        ln_likes = np.empty(values.shape)
        for ii, value in enumerate(values):
            non_marginalized.parameters[key] = value
            ln_likes[ii] = non_marginalized.log_likelihood_ratio()
        like = np.exp(ln_likes - max(ln_likes))

        marg_like = np.log(np.trapz(like * prior_values, values)) + max(ln_likes)
        self.assertAlmostEqual(
            marg_like, marginalized.log_likelihood_ratio(), delta=0.5
        )

    @parameterized.expand(_parameters)
    def test_marginalisation(self, kind, key, distance, time, phase):
        if all([distance, time, phase]):
            pytest.skip()
        tested_args = dict(
            distance_marginalization=distance,
            time_marginalization=time,
            phase_marginalization=phase,
        )
        marg_key = f"{key.split('_')[-1]}_marginalization"
        if tested_args[marg_key]:
            pytest.skip()
        reference_args = tested_args.copy()
        reference_args[marg_key] = True
        self._template(
            self.get_likelihood(kind, **reference_args),
            self.get_likelihood(kind, **tested_args),
            key=key,
        )

    def test_time_marginalisation_full_segment(self):
        """
        Test time marginalised likelihood matches brute force version over
        just part of a segment.
        """
        priors = self.priors.copy()
        prior = bilby.prior.Uniform(
            minimum=self.interferometers.start_time,
            maximum=self.interferometers.start_time + self.interferometers.duration,
        )
        priors["geocent_time"] = prior
        self._template(
            self.get_likelihood("regular", time_marginalization=True, priors=priors.copy()),
            self.get_likelihood("regular", priors=priors.copy()),
            key="geocent_time",
            values=self.waveform_generator.time_array,
            prior=prior,
        )

    @parameterized.expand(
        itertools.product(["regular", "roq"], *itertools.repeat([True, False], 3)),
        name_func=lambda func, num, param: (
            f"{func.__name__}_{num}__{param.args[0]}_" + "_".join([
                ["D", "P", "T"][ii] for ii, val
                in enumerate(param.args[1:]) if val
            ])
        )
    )
    def test_marginalization_reconstruction(self, kind, distance, phase, time):
        if time and kind == "roq":
            pytest.skip("Time reconstruction not supported for ROQ likelihood")
        marginalizations = dict(
            geocent_time=time,
            luminosity_distance=distance,
            phase=phase,
        )
        like = self.get_likelihood(
            kind=kind,
            distance_marginalization=distance,
            time_marginalization=time,
            phase_marginalization=phase,
        )
        params = self.parameters.copy()
        reference_values = dict(
            luminosity_distance=self.priors["luminosity_distance"].rescale(0.5),
            geocent_time=self.interferometers.start_time,
            phase=0.0,
        )
        for key in marginalizations:
            if marginalizations[key]:
                params[key] = reference_values[key]
        like.parameters.update(params)
        output = like.generate_posterior_sample_from_marginalized_likelihood()
        for key in marginalizations:
            self.assertFalse(marginalizations[key] and reference_values[key] == output[key])


class TestROQLikelihood(unittest.TestCase):
    def setUp(self):
        self.duration = 4
        self.sampling_frequency = 2048

        # Possible locations for the ROQ: in the docker image, local, or on CIT
        trial_roq_paths = [
            "/roq_basis",
            os.path.join(os.path.expanduser("~"), "ROQ_data/IMRPhenomPv2/4s"),
            "/home/cbc/ROQ_data/IMRPhenomPv2/4s",
        ]
        roq_dir = None
        for path in trial_roq_paths:
            if os.path.isdir(path):
                roq_dir = path
                break
        if roq_dir is None:
            raise Exception("Unable to load ROQ basis: cannot proceed with tests")

        linear_matrix_file = "{}/B_linear.npy".format(roq_dir)
        quadratic_matrix_file = "{}/B_quadratic.npy".format(roq_dir)

        fnodes_linear_file = "{}/fnodes_linear.npy".format(roq_dir)
        fnodes_linear = np.load(fnodes_linear_file).T
        fnodes_quadratic_file = "{}/fnodes_quadratic.npy".format(roq_dir)
        fnodes_quadratic = np.load(fnodes_quadratic_file).T
        self.linear_matrix_file = "{}/B_linear.npy".format(roq_dir)
        self.quadratic_matrix_file = "{}/B_quadratic.npy".format(roq_dir)
        self.params_file = "{}/params.dat".format(roq_dir)

        self.test_parameters = dict(
            mass_1=36.0,
            mass_2=36.0,
            a_1=0.0,
            a_2=0.0,
            tilt_1=0.0,
            tilt_2=0.0,
            phi_12=1.7,
            phi_jl=0.3,
            luminosity_distance=1000.0,
            theta_jn=0.4,
            psi=0.659,
            phase=1.3,
            geocent_time=1.2,
            ra=1.3,
            dec=-1.2,
        )

        ifos = bilby.gw.detector.InterferometerList(["H1"])
        ifos.set_strain_data_from_power_spectral_densities(
            sampling_frequency=self.sampling_frequency, duration=self.duration
        )

        self.priors = bilby.gw.prior.BBHPriorDict()
        self.priors.pop("mass_1")
        self.priors.pop("mass_2")
        # Testing is done with the 4s IMRPhenomPV2 ROQ basis
        self.priors["chirp_mass"] = bilby.core.prior.Uniform(12.299703, 45)
        self.priors["mass_ratio"] = bilby.core.prior.Uniform(0.125, 1)
        self.priors["geocent_time"] = bilby.core.prior.Uniform(1.19, 1.21)

        non_roq_wfg = bilby.gw.WaveformGenerator(
            duration=self.duration,
            sampling_frequency=self.sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
            waveform_arguments=dict(
                reference_frequency=20.0,
                minimum_frequency=20.0,
                approximant="IMRPhenomPv2",
            ),
        )

        ifos.inject_signal(
            parameters=self.test_parameters, waveform_generator=non_roq_wfg
        )

        self.ifos = ifos

        roq_wfg = bilby.gw.waveform_generator.WaveformGenerator(
            duration=self.duration,
            sampling_frequency=self.sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.binary_black_hole_roq,
            waveform_arguments=dict(
                frequency_nodes_linear=fnodes_linear,
                frequency_nodes_quadratic=fnodes_quadratic,
                reference_frequency=20.0,
                minimum_frequency=20.0,
                approximant="IMRPhenomPv2",
            ),
        )

        self.roq_wfg = roq_wfg

        self.non_roq = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=ifos, waveform_generator=non_roq_wfg
        )

        self.non_roq_phase = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=ifos,
            waveform_generator=non_roq_wfg,
            phase_marginalization=True,
            priors=self.priors.copy(),
        )

        self.roq = bilby.gw.likelihood.ROQGravitationalWaveTransient(
            interferometers=ifos,
            waveform_generator=roq_wfg,
            linear_matrix=linear_matrix_file,
            quadratic_matrix=quadratic_matrix_file,
            priors=self.priors,
        )

        self.roq_phase = bilby.gw.likelihood.ROQGravitationalWaveTransient(
            interferometers=ifos,
            waveform_generator=roq_wfg,
            linear_matrix=linear_matrix_file,
            quadratic_matrix=quadratic_matrix_file,
            phase_marginalization=True,
            priors=self.priors.copy(),
        )

    def tearDown(self):
        del (
            self.roq,
            self.non_roq,
            self.non_roq_phase,
            self.roq_phase,
            self.ifos,
            self.priors,
        )

    def test_matches_non_roq(self):
        self.non_roq.parameters.update(self.test_parameters)
        self.roq.parameters.update(self.test_parameters)
        self.assertLess(
            abs(self.non_roq.log_likelihood_ratio() - self.roq.log_likelihood_ratio())
            / self.non_roq.log_likelihood_ratio(),
            1e-3,
        )

    def test_time_prior_out_of_bounds_returns_zero(self):
        self.roq.parameters.update(self.test_parameters)
        self.roq.parameters["geocent_time"] = -5
        self.assertEqual(self.roq.log_likelihood_ratio(), np.nan_to_num(-np.inf))

    def test_phase_marginalisation_roq(self):
        """Test phase marginalised likelihood matches brute force version"""
        self.non_roq_phase.parameters = self.test_parameters.copy()
        self.roq_phase.parameters = self.test_parameters.copy()
        self.assertLess(
            abs(
                self.non_roq_phase.log_likelihood_ratio()
                - self.roq_phase.log_likelihood_ratio()
            )
            / self.non_roq_phase.log_likelihood_ratio(),
            1e-3,
        )

    def test_create_roq_weights_with_params(self):
        roq = bilby.gw.likelihood.ROQGravitationalWaveTransient(
            interferometers=self.ifos,
            waveform_generator=self.roq_wfg,
            linear_matrix=self.linear_matrix_file,
            roq_params=self.params_file,
            quadratic_matrix=self.quadratic_matrix_file,
            priors=self.priors,
        )
        roq.parameters.update(self.test_parameters)
        self.roq.parameters.update(self.test_parameters)
        self.assertEqual(roq.log_likelihood_ratio(), self.roq.log_likelihood_ratio())

    def test_create_roq_weights_frequency_mismatch_works_with_params(self):

        self.ifos[0].maximum_frequency = self.ifos[0].maximum_frequency / 2
        bilby.gw.likelihood.ROQGravitationalWaveTransient(
            interferometers=self.ifos,
            waveform_generator=self.roq_wfg,
            linear_matrix=self.linear_matrix_file,
            roq_params=self.params_file,
            quadratic_matrix=self.quadratic_matrix_file,
            priors=self.priors,
        )

    def test_create_roq_weights_frequency_mismatch_fails_without_params(self):
        self.ifos[0].maximum_frequency = self.ifos[0].maximum_frequency / 2
        with self.assertRaises(ValueError):
            bilby.gw.likelihood.ROQGravitationalWaveTransient(
                interferometers=self.ifos,
                waveform_generator=self.roq_wfg,
                linear_matrix=self.linear_matrix_file,
                quadratic_matrix=self.quadratic_matrix_file,
                priors=self.priors,
            )

    def test_create_roq_weights_fails_with_min_chirp_mass_outside_bounds(self):
        self.ifos[0].maximum_frequency = self.ifos[0].maximum_frequency / 2
        self.priors["chirp_mass"] = bilby.core.prior.Uniform(10, 45)
        with self.assertRaises(BilbyROQParamsRangeError):
            bilby.gw.likelihood.ROQGravitationalWaveTransient(
                interferometers=self.ifos,
                waveform_generator=self.roq_wfg,
                linear_matrix=self.linear_matrix_file,
                roq_params=self.params_file,
                quadratic_matrix=self.quadratic_matrix_file,
                priors=self.priors,
            )

    def test_create_roq_weights_fails_with_max_chirp_mass_outside_bounds(self):
        self.ifos[0].maximum_frequency = self.ifos[0].maximum_frequency / 2
        self.priors["chirp_mass"] = bilby.core.prior.Uniform(12.299703, 50)
        with self.assertRaises(BilbyROQParamsRangeError):
            bilby.gw.likelihood.ROQGravitationalWaveTransient(
                interferometers=self.ifos,
                waveform_generator=self.roq_wfg,
                linear_matrix=self.linear_matrix_file,
                roq_params=self.params_file,
                quadratic_matrix=self.quadratic_matrix_file,
                priors=self.priors,
            )

    def test_create_roq_weights_fails_with_min_component_mass_outside_bounds(self):
        self.ifos[0].maximum_frequency = self.ifos[0].maximum_frequency / 2
        self.priors["chirp_mass"] = bilby.core.prior.Uniform(12.299703, 45)
        self.priors["mass_ratio"] = bilby.core.prior.Uniform(1e-5, 1)
        with self.assertRaises(BilbyROQParamsRangeError):
            bilby.gw.likelihood.ROQGravitationalWaveTransient(
                interferometers=self.ifos,
                waveform_generator=self.roq_wfg,
                linear_matrix=self.linear_matrix_file,
                roq_params=self.params_file,
                quadratic_matrix=self.quadratic_matrix_file,
                priors=self.priors,
            )

    def test_create_roq_weights_fails_with_max_frequency(self):
        ifos = bilby.gw.detector.InterferometerList(["H1"])
        ifos.set_strain_data_from_power_spectral_densities(
            sampling_frequency=2 ** 14, duration=4
        )
        ifos[0].maximum_frequency = 2 ** 13
        with self.assertRaises(BilbyROQParamsRangeError):
            bilby.gw.likelihood.ROQGravitationalWaveTransient(
                interferometers=ifos,
                waveform_generator=self.roq_wfg,
                linear_matrix=self.linear_matrix_file,
                roq_params=self.params_file,
                quadratic_matrix=self.quadratic_matrix_file,
                priors=self.priors,
            )

    def test_create_roq_weights_fails_due_to_min_frequency(self):
        self.ifos[0].minimum_frequency = 15
        with self.assertRaises(BilbyROQParamsRangeError):
            bilby.gw.likelihood.ROQGravitationalWaveTransient(
                interferometers=self.ifos,
                waveform_generator=self.roq_wfg,
                linear_matrix=self.linear_matrix_file,
                roq_params=self.params_file,
                quadratic_matrix=self.quadratic_matrix_file,
                priors=self.priors,
            )

    def test_create_roq_weights_fails_due_to_duration(self):
        ifos = bilby.gw.detector.InterferometerList(["H1"])
        ifos.set_strain_data_from_power_spectral_densities(
            sampling_frequency=self.sampling_frequency, duration=16
        )
        with self.assertRaises(BilbyROQParamsRangeError):
            bilby.gw.likelihood.ROQGravitationalWaveTransient(
                interferometers=ifos,
                waveform_generator=self.roq_wfg,
                linear_matrix=self.linear_matrix_file,
                roq_params=self.params_file,
                quadratic_matrix=self.quadratic_matrix_file,
                priors=self.priors,
            )


class TestRescaledROQLikelihood(unittest.TestCase):
    def test_rescaling(self):

        # Possible locations for the ROQ: in the docker image, local, or on CIT
        trial_roq_paths = [
            "/roq_basis",
            os.path.join(os.path.expanduser("~"), "ROQ_data/IMRPhenomPv2/4s"),
            "/home/cbc/ROQ_data/IMRPhenomPv2/4s",
        ]
        roq_dir = None
        for path in trial_roq_paths:
            if os.path.isdir(path):
                roq_dir = path
                break
        if roq_dir is None:
            raise Exception("Unable to load ROQ basis: cannot proceed with tests")

        linear_matrix_file = "{}/B_linear.npy".format(roq_dir)
        quadratic_matrix_file = "{}/B_quadratic.npy".format(roq_dir)

        fnodes_linear_file = "{}/fnodes_linear.npy".format(roq_dir)
        fnodes_linear = np.load(fnodes_linear_file).T
        fnodes_quadratic_file = "{}/fnodes_quadratic.npy".format(roq_dir)
        fnodes_quadratic = np.load(fnodes_quadratic_file).T
        self.linear_matrix_file = "{}/B_linear.npy".format(roq_dir)
        self.quadratic_matrix_file = "{}/B_quadratic.npy".format(roq_dir)
        self.params_file = "{}/params.dat".format(roq_dir)

        scale_factor = 0.5
        params = np.genfromtxt(self.params_file, names=True)

        self.duration = 4 / scale_factor
        self.sampling_frequency = 2048 * scale_factor

        ifos = bilby.gw.detector.InterferometerList(["H1"])
        ifos.set_strain_data_from_power_spectral_densities(
            sampling_frequency=self.sampling_frequency, duration=self.duration
        )
        self.ifos = ifos

        self.priors = bilby.gw.prior.BBHPriorDict()
        self.priors.pop("mass_1")
        self.priors.pop("mass_2")
        # Testing is done with the 4s IMRPhenomPV2 ROQ basis
        self.priors["chirp_mass"] = bilby.core.prior.Uniform(
            12.299703 / scale_factor, 45 / scale_factor
        )
        self.priors["mass_ratio"] = bilby.core.prior.Uniform(0.125, 1)
        self.priors["geocent_time"] = bilby.core.prior.Uniform(1.19, 1.21)

        self.roq_wfg = bilby.gw.waveform_generator.WaveformGenerator(
            duration=self.duration,
            sampling_frequency=self.sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.binary_black_hole_roq,
            waveform_arguments=dict(
                frequency_nodes_linear=fnodes_linear,
                frequency_nodes_quadratic=fnodes_quadratic,
                reference_frequency=20.0,
                minimum_frequency=20.0,
                approximant="IMRPhenomPv2",
            ),
        )

        self.roq = bilby.gw.likelihood.ROQGravitationalWaveTransient(
            interferometers=ifos,
            waveform_generator=self.roq_wfg,
            linear_matrix=linear_matrix_file,
            roq_params=params,
            roq_scale_factor=scale_factor,
            quadratic_matrix=quadratic_matrix_file,
            priors=self.priors,
        )


class TestROQLikelihoodHDF5(unittest.TestCase):
    """
    Test ROQ likelihood constructed from .hdf5 basis

    The .hdf5 files contain 3 linear bases constructed over 8Msun<Mc<10Msun, 10Msun<Mc<12Msun, and 12Msun<Mc<14Msun
    respectively, and 2 quadratic bases constructed over 8Msun<Mc<11Msun and 11Msun<Mc<14Msun respectively.

    """

    _path_to_basis = "/roq_basis/basis_addcal.hdf5"
    _path_to_basis_mb = "/roq_basis/basis_multiband_addcal.hdf5"

    def setUp(self):
        self.minimum_frequency = 20
        self.sampling_frequency = 2048
        self.duration = 16
        self.reference_frequency = 20.0
        self.waveform_approximant = "IMRPhenomD"
        # The SNRs of injections are 130-160 for roq_scale_factor=1 and 70-80 for roq_scale_factor=2
        self.injection_parameters = dict(
            mass_ratio=0.8,
            chi_1=0.0,
            chi_2=0.0,
            luminosity_distance=100.0,
            theta_jn=0.4,
            psi=0.659,
            phase=1.3,
            geocent_time=1.2,
            ra=1.3,
            dec=-1.2
        )
        self.priors = bilby.gw.prior.BBHPriorDict()
        self.priors.pop("mass_1")
        self.priors.pop("mass_2")
        self.priors["mass_ratio"] = bilby.core.prior.Uniform(0.125, 1)
        self.priors["geocent_time"] = bilby.core.prior.Uniform(
            self.injection_parameters["geocent_time"] - 0.1,
            self.injection_parameters["geocent_time"] + 0.1
        )

    @parameterized.expand(
        product(
            [_path_to_basis, _path_to_basis_mb],
            [_path_to_basis, _path_to_basis_mb],
            [(8, 9), (8, 10.5), (8, 11.5), (8, 12.5), (8, 14)],
            [1, 2]
        )
    )
    def test_number_of_loaded_bases(self, basis_linear, basis_quadratic, mc_range, roq_scale_factor):
        "Check if ROQ weights are computed only for the bases in the prior range"
        self.minimum_frequency *= roq_scale_factor
        self.sampling_frequency *= roq_scale_factor
        self.duration /= roq_scale_factor
        self.reference_frequency *= roq_scale_factor
        mc_min, mc_max = mc_range
        mc_min /= roq_scale_factor
        mc_max /= roq_scale_factor
        self.priors["chirp_mass"].minimum = mc_min
        self.priors["chirp_mass"].maximum = mc_max

        interferometers = bilby.gw.detector.InterferometerList(["H1", "L1"])
        interferometers.set_strain_data_from_power_spectral_densities(
            sampling_frequency=self.sampling_frequency,
            duration=self.duration,
            start_time=self.injection_parameters["geocent_time"] - self.duration + 1
        )
        for ifo in interferometers:
            ifo.minimum_frequency = self.minimum_frequency

        search_waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
            duration=self.duration,
            sampling_frequency=self.sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.binary_black_hole_roq,
            waveform_arguments=dict(
                reference_frequency=self.reference_frequency,
                waveform_approximant=self.waveform_approximant
            )
        )

        likelihood = bilby.gw.likelihood.ROQGravitationalWaveTransient(
            interferometers=interferometers,
            priors=self.priors,
            waveform_generator=search_waveform_generator,
            linear_matrix=basis_linear,
            quadratic_matrix=basis_quadratic,
            roq_scale_factor=roq_scale_factor
        )

        with h5py.File(basis_linear, "r") as f:
            mc_ranges_linear = f["prior_range_linear"]["chirp_mass"][()] / roq_scale_factor
        with h5py.File(basis_quadratic, "r") as f:
            mc_ranges_quadratic = f["prior_range_quadratic"]["chirp_mass"][()] / roq_scale_factor
        number_of_bases_linear = np.sum(
            (mc_ranges_linear[:, 1] >= self.priors["chirp_mass"].minimum) *
            (mc_ranges_linear[:, 0] <= self.priors["chirp_mass"].maximum)
        )
        number_of_bases_quadratic = np.sum(
            (mc_ranges_quadratic[:, 1] >= self.priors["chirp_mass"].minimum) *
            (mc_ranges_quadratic[:, 0] <= self.priors["chirp_mass"].maximum)
        )

        self.assertEqual(likelihood.number_of_bases_linear, number_of_bases_linear)
        self.assertEqual(likelihood.number_of_bases_quadratic, number_of_bases_quadratic)
        self.assertEqual(len(likelihood.weights['frequency_nodes_linear']), number_of_bases_linear)
        self.assertEqual(len(likelihood.weights['frequency_nodes_quadratic']), number_of_bases_quadratic)
        for ifo in interferometers:
            self.assertEqual(len(likelihood.weights['{}_linear'.format(ifo.name)]), number_of_bases_linear)
            self.assertEqual(len(likelihood.weights['{}_quadratic'.format(ifo.name)]), number_of_bases_quadratic)

    @parameterized.expand(
        product(
            [_path_to_basis, _path_to_basis_mb],
            [_path_to_basis, _path_to_basis_mb],
            [(8, 9), (8, 14)],
            [1, 2],
            [False, True]
        )
    )
    def test_likelihood_accuracy(self, basis_linear, basis_quadratic, mc_range, roq_scale_factor, add_cal_errors):
        "Compare with log likelihood ratios computed by the non-ROQ likelihood"
        self.minimum_frequency *= roq_scale_factor
        self.sampling_frequency *= roq_scale_factor
        self.duration /= roq_scale_factor
        self.reference_frequency *= roq_scale_factor
        mc_min, mc_max = mc_range
        mc_min /= roq_scale_factor
        mc_max /= roq_scale_factor
        self.injection_parameters["chirp_mass"] = (mc_min + mc_max) / 2
        self.priors["chirp_mass"].minimum = mc_min
        self.priors["chirp_mass"].maximum = mc_max

        interferometers = bilby.gw.detector.InterferometerList(["H1", "L1"])
        for ifo in interferometers:
            ifo.minimum_frequency = self.minimum_frequency
        interferometers.set_strain_data_from_zero_noise(
            sampling_frequency=self.sampling_frequency,
            duration=self.duration,
            start_time=self.injection_parameters["geocent_time"] - self.duration + 1
        )

        if add_cal_errors:
            spline_calibration_nodes = 10
            np.random.seed(170817)
            for ifo in interferometers:
                prefix = f"recalib_{ifo.name}_"
                ifo.calibration_model = bilby.gw.calibration.CubicSpline(
                    prefix=prefix,
                    minimum_frequency=ifo.minimum_frequency,
                    maximum_frequency=ifo.maximum_frequency,
                    n_points=spline_calibration_nodes
                )
                for i in range(spline_calibration_nodes):
                    # 5% in amplitude, 5deg in phase
                    self.injection_parameters[f"{prefix}amplitude_{i}"] = \
                        np.random.normal(loc=0, scale=0.05)
                    self.injection_parameters[f"{prefix}phase_{i}"] = \
                        np.random.normal(loc=0, scale=5 * np.pi / 180)

        waveform_generator = bilby.gw.WaveformGenerator(
            duration=self.duration,
            sampling_frequency=self.sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
            waveform_arguments=dict(
                reference_frequency=self.reference_frequency,
                waveform_approximant=self.waveform_approximant
            )
        )
        interferometers.inject_signal(waveform_generator=waveform_generator, parameters=self.injection_parameters)

        likelihood = bilby.gw.GravitationalWaveTransient(
            interferometers=interferometers,
            waveform_generator=waveform_generator,
            priors=self.priors
        )

        search_waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
            duration=self.duration,
            sampling_frequency=self.sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.binary_black_hole_roq,
            waveform_arguments=dict(
                reference_frequency=self.reference_frequency,
                waveform_approximant=self.waveform_approximant
            )
        )
        likelihood_roq = bilby.gw.likelihood.ROQGravitationalWaveTransient(
            interferometers=interferometers,
            priors=self.priors,
            waveform_generator=search_waveform_generator,
            linear_matrix=basis_linear,
            quadratic_matrix=basis_quadratic,
            roq_scale_factor=roq_scale_factor
        )
        # The maximum error of log likelihood ratio. It is set to be larger for roq_scale_factor=1 as the injected SNR
        # is higher.
        if roq_scale_factor == 1:
            max_llr_error = 5e-1
        elif roq_scale_factor == 2:
            max_llr_error = 5e-2
        else:
            raise
        for mc in np.linspace(self.priors["chirp_mass"].minimum, self.priors["chirp_mass"].maximum, 11):
            parameters = self.injection_parameters.copy()
            parameters["chirp_mass"] = mc
            likelihood.parameters.update(parameters)
            likelihood_roq.parameters.update(parameters)
            llr = likelihood.log_likelihood_ratio()
            llr_roq = likelihood_roq.log_likelihood_ratio()
            self.assertLess(np.abs(llr - llr_roq), max_llr_error)


class TestCreateROQLikelihood(unittest.TestCase):
    """
    Test if ROQ likelihood is constructed without any errors from .hdf5 or .npy basis

    The .hdf5 files contain 3 linear bases constructed over 8Msun<Mc<10Msun, 10Msun<Mc<12Msun, and 12Msun<Mc<14Msun
    respectively, and 2 quadratic bases constructed over 8Msun<Mc<11Msun and 11Msun<Mc<14Msun respectively.

    """

    _path_to_basis = "/roq_basis/basis_addcal.hdf5"
    _path_to_basis_mb = "/roq_basis/basis_multiband_addcal.hdf5"

    @parameterized.expand(product([_path_to_basis, _path_to_basis_mb], [_path_to_basis, _path_to_basis_mb]))
    def test_from_hdf5(self, basis_linear, basis_quadratic):
        minimum_frequency = 20
        sampling_frequency = 2048
        duration = 16
        geocent_time = 1.2
        reference_frequency = 20.0
        waveform_approximant = "IMRPhenomD"
        mc_range = [8, 14]

        priors = bilby.gw.prior.BBHPriorDict()
        priors["geocent_time"] = bilby.core.prior.Uniform(geocent_time - 0.1, geocent_time + 0.1)
        priors["chirp_mass"].minimum = mc_range[0]
        priors["chirp_mass"].maximum = mc_range[1]

        interferometers = bilby.gw.detector.InterferometerList(["H1", "L1"])
        interferometers.set_strain_data_from_power_spectral_densities(
            sampling_frequency=sampling_frequency, duration=duration, start_time=geocent_time - duration + 1
        )
        for ifo in interferometers:
            ifo.minimum_frequency = minimum_frequency

        search_waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
            duration=duration,
            sampling_frequency=sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.binary_black_hole_roq,
            waveform_arguments=dict(
                reference_frequency=reference_frequency,
                waveform_approximant=waveform_approximant
            )
        )

        bilby.gw.likelihood.ROQGravitationalWaveTransient(
            interferometers=interferometers,
            priors=priors,
            waveform_generator=search_waveform_generator,
            linear_matrix=basis_linear,
            quadratic_matrix=basis_quadratic
        )

    @parameterized.expand([(False, ), (True, )])
    def test_from_npy(self, from_array):
        # Possible locations for the ROQ: in the docker image, local, or on CIT
        trial_roq_paths = [
            "/roq_basis",
            os.path.join(os.path.expanduser("~"), "ROQ_data/IMRPhenomPv2/4s"),
            "/home/cbc/ROQ_data/IMRPhenomPv2/4s",
        ]
        roq_dir = None
        for path in trial_roq_paths:
            if os.path.isdir(path):
                roq_dir = path
                break
        if roq_dir is None:
            raise Exception("Unable to load ROQ basis: cannot proceed with tests")

        basis_linear = "{}/B_linear.npy".format(roq_dir)
        if from_array:
            basis_linear = np.load(basis_linear).T
        basis_quadratic = "{}/B_quadratic.npy".format(roq_dir)
        if from_array:
            basis_quadratic = np.load(basis_quadratic).T
        fnodes_linear = np.load("{}/fnodes_linear.npy".format(roq_dir))
        fnodes_quadratic = np.load("{}/fnodes_quadratic.npy".format(roq_dir))
        params_file = "{}/params.dat".format(roq_dir)

        minimum_frequency = 20
        sampling_frequency = 2048
        duration = 4
        geocent_time = 1.2
        reference_frequency = 20.0
        waveform_approximant = "IMRPhenomPv2"
        mc_range = [12.299703, 45]

        priors = bilby.gw.prior.BBHPriorDict()
        priors["geocent_time"] = bilby.core.prior.Uniform(geocent_time - 0.1, geocent_time + 0.1)
        priors["chirp_mass"].minimum = mc_range[0]
        priors["chirp_mass"].maximum = mc_range[1]

        interferometers = bilby.gw.detector.InterferometerList(["H1", "L1"])
        interferometers.set_strain_data_from_power_spectral_densities(
            sampling_frequency=sampling_frequency, duration=duration, start_time=geocent_time - duration + 1
        )
        for ifo in interferometers:
            ifo.minimum_frequency = minimum_frequency

        search_waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
            duration=duration,
            sampling_frequency=sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.binary_black_hole_roq,
            waveform_arguments=dict(
                frequency_nodes_linear=fnodes_linear,
                frequency_nodes_quadratic=fnodes_quadratic,
                reference_frequency=reference_frequency,
                waveform_approximant=waveform_approximant
            )
        )

        bilby.gw.likelihood.ROQGravitationalWaveTransient(
            interferometers=interferometers,
            priors=priors,
            waveform_generator=search_waveform_generator,
            linear_matrix=basis_linear,
            quadratic_matrix=basis_quadratic,
            roq_params=params_file
        )


class TestInOutROQWeights(unittest.TestCase):

    @parameterized.expand(['npz', 'json', 'hdf5'])
    def test_out_single_basis(self, format):
        likelihood = self.create_likelihood_single_basis()
        filename = f'weights.{format}'
        likelihood.save_weights(filename, format=format)
        self.assertTrue(os.path.exists(filename))

    @parameterized.expand(['npz', 'json', 'hdf5'])
    def test_in_single_basis(self, format):
        likelihood = self.create_likelihood_single_basis()
        filename = f'weights.{format}'
        likelihood.save_weights(filename, format=format)
        likelihood_from_weights = bilby.gw.likelihood.ROQGravitationalWaveTransient(
            interferometers=likelihood.interferometers,
            priors=likelihood.priors,
            waveform_generator=likelihood.waveform_generator,
            weights=filename
        )
        self.check_weights_are_same(likelihood, likelihood_from_weights)

    @parameterized.expand([(False, ), (True, )])
    def test_out_multiple_bases(self, multiband):
        format = 'hdf5'
        filename = f'weights.{format}'
        likelihood = self.create_likelihood_multiple_bases(multiband)
        likelihood.save_weights(filename, format=format)
        self.assertTrue(os.path.exists(filename))

    @parameterized.expand([(False, ), (True, )])
    def test_in_multiple_bases(self, multiband):
        format = 'hdf5'
        filename = f'weights.{format}'
        likelihood = self.create_likelihood_multiple_bases(multiband)
        likelihood.save_weights(filename, format=format)
        likelihood_from_weights = bilby.gw.likelihood.ROQGravitationalWaveTransient(
            interferometers=likelihood.interferometers,
            priors=likelihood.priors,
            waveform_generator=likelihood.waveform_generator,
            weights=filename
        )
        self.check_weights_are_same(likelihood, likelihood_from_weights)

    @parameterized.expand(product(['npz', 'json'], [False, True]))
    def test_out_multiple_bases_inconsistent_format(self, format, multiband):
        "npz or json format is not compatible with multiple bases"
        likelihood = self.create_likelihood_multiple_bases(multiband)
        with self.assertRaises(ValueError):
            likelihood.save_weights('weights', format=format)

    def tearDown(self):
        for format in ['npz', 'json', 'hdf5']:
            filename = f'weights.{format}'
            if os.path.exists(filename):
                os.remove(filename)

    @staticmethod
    def check_weights_are_same(l1, l2):
        """Check if input likelihoods contain same ROQ weights

        Parameters
        ==========
        l1, l2: bilby.gw.likelihood.ROQGravitationalWaveTransient

        """
        np.testing.assert_array_almost_equal(l1.weights['time_samples'], l2.weights['time_samples'])
        for basis_type in ['linear', 'quadratic']:
            # check weights
            for ifo in l1.interferometers:
                key = f'{ifo.name}_{basis_type}'
                for i in range(len(l1.weights[key])):
                    np.testing.assert_array_almost_equal(l1.weights[key][i], l2.weights[key][i])
            # check prior ranges
            key = f'prior_range_{basis_type}'
            if key in l1.weights:
                for param_name in l1.weights[key]:
                    np.testing.assert_array_almost_equal(l1.weights[key][param_name], l2.weights[key][param_name])
            # check frequency nodes
            key = f'frequency_nodes_{basis_type}'
            if key in l1.weights:
                for i in range(len(l1.weights[key])):
                    np.testing.assert_array_almost_equal(l1.weights[key][i], l2.weights[key][i])

    def create_likelihood_single_basis(self):
        # Possible locations for the ROQ: in the docker image, local, or on CIT
        trial_roq_paths = [
            "/roq_basis",
            os.path.join(os.path.expanduser("~"), "ROQ_data/IMRPhenomPv2/4s"),
            "/home/cbc/ROQ_data/IMRPhenomPv2/4s",
        ]
        roq_dir = None
        for path in trial_roq_paths:
            if os.path.isdir(path):
                roq_dir = path
                break
        if roq_dir is None:
            raise Exception("Unable to load ROQ basis: cannot proceed with tests")

        linear_matrix_file = "{}/B_linear.npy".format(roq_dir)
        quadratic_matrix_file = "{}/B_quadratic.npy".format(roq_dir)
        fnodes_linear = np.load("{}/fnodes_linear.npy".format(roq_dir))
        fnodes_quadratic = np.load("{}/fnodes_quadratic.npy".format(roq_dir))

        minimum_frequency = 20
        sampling_frequency = 2048
        duration = 4
        geocent_time = 1.2
        reference_frequency = 20.0
        waveform_approximant = "IMRPhenomPv2"
        mc_range = [12.299703, 45]

        priors = bilby.gw.prior.BBHPriorDict()
        priors["geocent_time"] = bilby.core.prior.Uniform(geocent_time - 0.001, geocent_time + 0.001)
        priors["chirp_mass"].minimum = mc_range[0]
        priors["chirp_mass"].maximum = mc_range[1]

        interferometers = bilby.gw.detector.InterferometerList(["H1", "L1"])
        interferometers.set_strain_data_from_power_spectral_densities(
            sampling_frequency=sampling_frequency, duration=duration, start_time=geocent_time - duration + 1
        )
        for ifo in interferometers:
            ifo.minimum_frequency = minimum_frequency

        search_waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
            duration=duration,
            sampling_frequency=sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.binary_black_hole_roq,
            waveform_arguments=dict(
                frequency_nodes_linear=fnodes_linear,
                frequency_nodes_quadratic=fnodes_quadratic,
                reference_frequency=reference_frequency,
                waveform_approximant=waveform_approximant
            )
        )

        return bilby.gw.likelihood.ROQGravitationalWaveTransient(
            interferometers=interferometers,
            priors=priors,
            waveform_generator=search_waveform_generator,
            linear_matrix=linear_matrix_file,
            quadratic_matrix=quadratic_matrix_file
        )

    def create_likelihood_multiple_bases(self, multiband):
        minimum_frequency = 20
        sampling_frequency = 2048
        duration = 16
        geocent_time = 1.2
        reference_frequency = 20.0
        waveform_approximant = "IMRPhenomD"
        mc_range = [8, 14]

        priors = bilby.gw.prior.BBHPriorDict()
        priors["geocent_time"] = bilby.core.prior.Uniform(geocent_time - 0.001, geocent_time + 0.001)
        priors["chirp_mass"].minimum = mc_range[0]
        priors["chirp_mass"].maximum = mc_range[1]

        interferometers = bilby.gw.detector.InterferometerList(["H1", "L1"])
        interferometers.set_strain_data_from_power_spectral_densities(
            sampling_frequency=sampling_frequency, duration=duration, start_time=geocent_time - duration + 1
        )
        for ifo in interferometers:
            ifo.minimum_frequency = minimum_frequency

        search_waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
            duration=duration,
            sampling_frequency=sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.binary_black_hole_roq,
            waveform_arguments=dict(
                reference_frequency=reference_frequency,
                waveform_approximant=waveform_approximant
            )
        )

        if multiband:
            path_to_basis = "/roq_basis/basis_multiband_addcal.hdf5"
        else:
            path_to_basis = "/roq_basis/basis_addcal.hdf5"
        return bilby.gw.likelihood.ROQGravitationalWaveTransient(
            interferometers=interferometers,
            priors=priors,
            waveform_generator=search_waveform_generator,
            linear_matrix=path_to_basis,
            quadratic_matrix=path_to_basis
        )


class TestBBHLikelihoodSetUp(unittest.TestCase):
    def setUp(self):
        self.ifos = bilby.gw.detector.InterferometerList(["H1"])

    def tearDown(self):
        del self.ifos

    def test_instantiation(self):
        self.like = bilby.gw.likelihood.get_binary_black_hole_likelihood(self.ifos)


class TestMBLikelihood(unittest.TestCase):
    def setUp(self):
        duration = 16
        fmin = 20.
        sampling_frequency = 2048.
        self.test_parameters = dict(
            chirp_mass=6.0,
            mass_ratio=0.5,
            a_1=0.0,
            a_2=0.0,
            tilt_1=0.0,
            tilt_2=0.0,
            phi_12=0.0,
            phi_jl=0.0,
            luminosity_distance=200.0,
            theta_jn=0.4,
            psi=0.659,
            phase=1.3,
            geocent_time=1187008882,
            ra=1.3,
            dec=-1.2
        )  # Network SNR is ~50

        ifos = bilby.gw.detector.InterferometerList(["H1", "L1", "V1"])
        np.random.seed(170817)
        ifos.set_strain_data_from_power_spectral_densities(
            sampling_frequency=sampling_frequency, duration=duration,
            start_time=self.test_parameters['geocent_time'] - duration + 2.
        )
        for ifo in ifos:
            ifo.minimum_frequency = fmin

        spline_calibration_nodes = 10
        self.calibration_parameters = {}
        for ifo in ifos:
            ifo.calibration_model = bilby.gw.calibration.CubicSpline(
                prefix=f"recalib_{ifo.name}_",
                minimum_frequency=ifo.minimum_frequency,
                maximum_frequency=ifo.maximum_frequency,
                n_points=spline_calibration_nodes
            )
            for i in range(spline_calibration_nodes):
                self.test_parameters[f"recalib_{ifo.name}_amplitude_{i}"] = 0
                self.test_parameters[f"recalib_{ifo.name}_phase_{i}"] = 0
                # Calibration errors of 5% in amplitude and 5 degrees in phase
                self.calibration_parameters[f"recalib_{ifo.name}_amplitude_{i}"] = \
                    np.random.normal(loc=0, scale=0.05)
                self.calibration_parameters[f"recalib_{ifo.name}_phase_{i}"] = \
                    np.random.normal(loc=0, scale=5 * np.pi / 180)

        priors = bilby.gw.prior.BBHPriorDict()
        priors.pop("mass_1")
        priors.pop("mass_2")
        priors["chirp_mass"] = bilby.core.prior.Uniform(5.5, 6.5)
        priors["mass_ratio"] = bilby.core.prior.Uniform(0.125, 1)
        priors["geocent_time"] = bilby.core.prior.Uniform(
            self.test_parameters['geocent_time'] - 0.1,
            self.test_parameters['geocent_time'] + 0.1)

        approximant_22 = "IMRPhenomD"
        approximant_homs = "IMRPhenomHM"
        non_mb_wfg_22 = bilby.gw.WaveformGenerator(
            duration=duration, sampling_frequency=sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
            waveform_arguments=dict(
                reference_frequency=fmin, minimum_frequency=fmin, approximant=approximant_22)
        )
        mb_wfg_22 = bilby.gw.waveform_generator.WaveformGenerator(
            duration=duration, sampling_frequency=sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.binary_black_hole_frequency_sequence,
            waveform_arguments=dict(
                reference_frequency=fmin, approximant=approximant_22)
        )
        non_mb_wfg_homs = bilby.gw.WaveformGenerator(
            duration=duration, sampling_frequency=sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
            waveform_arguments=dict(
                reference_frequency=fmin, minimum_frequency=fmin, approximant=approximant_homs)
        )
        mb_wfg_homs = bilby.gw.waveform_generator.WaveformGenerator(
            duration=duration, sampling_frequency=sampling_frequency,
            frequency_domain_source_model=bilby.gw.source.binary_black_hole_frequency_sequence,
            waveform_arguments=dict(
                reference_frequency=fmin, approximant=approximant_homs)
        )

        ifos_22 = deepcopy(ifos)
        ifos_22.inject_signal(
            parameters=self.test_parameters, waveform_generator=non_mb_wfg_22
        )
        ifos_homs = deepcopy(ifos)
        ifos_homs.inject_signal(
            parameters=self.test_parameters, waveform_generator=non_mb_wfg_homs
        )

        self.non_mb_22 = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=ifos_22, waveform_generator=non_mb_wfg_22
        )
        self.non_mb_homs = bilby.gw.likelihood.GravitationalWaveTransient(
            interferometers=ifos_homs, waveform_generator=non_mb_wfg_homs
        )

        self.mb_22 = bilby.gw.likelihood.MBGravitationalWaveTransient(
            interferometers=ifos_22, waveform_generator=deepcopy(mb_wfg_22),
            reference_chirp_mass=self.test_parameters['chirp_mass'],
            priors=priors.copy()
        )
        self.mb_ifftfft_22 = bilby.gw.likelihood.MBGravitationalWaveTransient(
            interferometers=ifos_22, waveform_generator=deepcopy(mb_wfg_22),
            reference_chirp_mass=self.test_parameters['chirp_mass'],
            priors=priors.copy(), linear_interpolation=False
        )
        self.mb_homs = bilby.gw.likelihood.MBGravitationalWaveTransient(
            interferometers=ifos_homs, waveform_generator=deepcopy(mb_wfg_homs),
            reference_chirp_mass=self.test_parameters['chirp_mass'],
            priors=priors.copy(), linear_interpolation=False, highest_mode=4
        )
        self.mb_more_accurate = bilby.gw.likelihood.MBGravitationalWaveTransient(
            interferometers=ifos_22, waveform_generator=deepcopy(mb_wfg_22),
            reference_chirp_mass=self.test_parameters['chirp_mass'],
            priors=priors.copy(), accuracy_factor=50
        )

    def tearDown(self):
        del (
            self.non_mb_22,
            self.non_mb_homs,
            self.mb_22,
            self.mb_ifftfft_22,
            self.mb_homs,
            self.mb_more_accurate
        )

    @parameterized.expand([(False, ), (True, )])
    def test_matches_non_mb(self, add_cal_errors):
        self.non_mb_22.parameters.update(self.test_parameters)
        self.mb_22.parameters.update(self.test_parameters)
        if add_cal_errors:
            self.non_mb_22.parameters.update(self.calibration_parameters)
            self.mb_22.parameters.update(self.calibration_parameters)
        self.assertLess(
            abs(self.non_mb_22.log_likelihood_ratio() - self.mb_22.log_likelihood_ratio()),
            1.5e-2
        )

    @parameterized.expand([(False, ), (True, )])
    def test_ifft_fft(self, add_cal_errors):
        """
        Check if multi-banding likelihood with (h, h) computed with the
        IFFT-FFT algorithm matches the original likelihood.
        """
        self.non_mb_22.parameters.update(self.test_parameters)
        self.mb_ifftfft_22.parameters.update(self.test_parameters)
        if add_cal_errors:
            self.non_mb_22.parameters.update(self.calibration_parameters)
            self.mb_ifftfft_22.parameters.update(self.calibration_parameters)
        self.assertLess(
            abs(self.non_mb_22.log_likelihood_ratio() - self.mb_ifftfft_22.log_likelihood_ratio()),
            6e-3
        )

    @parameterized.expand([(False, ), (True, )])
    def test_homs(self, add_cal_errors):
        """
        Check if multi-banding likelihood matches the original likelihood for higher-order moments.
        """
        self.non_mb_homs.parameters.update(self.test_parameters)
        self.mb_homs.parameters.update(self.test_parameters)
        if add_cal_errors:
            self.non_mb_homs.parameters.update(self.calibration_parameters)
            self.mb_homs.parameters.update(self.calibration_parameters)
        self.assertLess(
            abs(self.non_mb_homs.log_likelihood_ratio() - self.mb_homs.log_likelihood_ratio()),
            1e-3
        )

    def test_large_accuracy_factor(self):
        """
        Check if larger accuracy factor increases the accuracy.
        """
        self.non_mb_22.parameters.update(self.test_parameters)
        self.mb_22.parameters.update(self.test_parameters)
        self.mb_more_accurate.parameters.update(self.test_parameters)
        self.assertLess(
            abs(self.non_mb_22.log_likelihood_ratio() - self.mb_more_accurate.log_likelihood_ratio()),
            abs(self.non_mb_22.log_likelihood_ratio() - self.mb_22.log_likelihood_ratio()) / 2
        )


if __name__ == "__main__":
    unittest.main()



import os
import copy
import shutil
import unittest
import inspect
import importlib
import sys
import time
import bilby
from bilby.bilby_mcmc.chain import Chain, Sample
from bilby.bilby_mcmc import proposals
from bilby.bilby_mcmc.utils import LOGLKEY, LOGPKEY
import numpy as np


class GivenProposal(proposals.BaseProposal):
    """ A simple proposal class used for testing """
    def __init__(self, priors, weight=1, subset=None, sigma=0.01):
        super(GivenProposal, self).__init__(priors, weight, subset)

    def propose(self, chain):
        log_factor = 0
        return self.given_sample, log_factor


class TestBaseProposals(unittest.TestCase):
    def create_priors(self, ndim=2, boundary=None):
        priors = bilby.core.prior.PriorDict({
            f'x{i}': bilby.core.prior.Uniform(-10, 10, name=f'x{i}', boundary=boundary)
            for i in range(ndim)
        })
        priors["fixedA"] = bilby.core.prior.DeltaFunction(1)
        priors["infinite_support"] = bilby.core.prior.Normal(0, 1)
        priors["half_infinite_support"] = bilby.core.prior.HalfNormal(1)
        return priors

    def create_random_sample(self, ndim=2):
        p = {f"x{i}": np.random.normal(0, 1) for i in range(ndim)}
        p[LOGLKEY] = np.random.normal(0, 1)
        p[LOGPKEY] = -1
        p["fixedA"] = 1
        p["infinite_support"] = np.random.normal(0, 1)
        p["half_infinite_support"] = np.abs(np.random.normal(0, 1))
        return Sample(p)

    def create_chain(self, n=1000, ndim=2):
        initial_sample = self.create_random_sample(ndim)
        chain = Chain(initial_sample=initial_sample)
        for i in range(n):
            chain.append(self.create_random_sample(ndim))
        return chain

    def test_GivenProposal(self):
        priors = self.create_priors()
        chain = self.create_chain()
        proposal = GivenProposal(priors)
        proposal.given_sample = self.create_random_sample()
        prop, _ = proposal(chain)
        self.assertEqual(prop, proposal.given_sample)

    def test_noboundary(self):
        priors = self.create_priors()
        chain = self.create_chain()
        proposal = GivenProposal(priors)

        sample = self.create_random_sample()
        sample["x0"] = priors["x0"].maximum + 0.5
        proposal.given_sample = sample

        prop, _ = proposal(chain)
        self.assertEqual(prop, proposal.given_sample)
        self.assertEqual(prop["x0"], priors["x0"].maximum + 0.5)

    def test_periodic_boundary_above(self):
        priors = self.create_priors(boundary="periodic")
        chain = self.create_chain()
        proposal = GivenProposal(priors)

        sample = self.create_random_sample()
        sample["x0"] = priors["x0"].maximum + 0.5
        proposal.given_sample = copy.deepcopy(sample)

        prop, _ = proposal(chain)
        self.assertFalse(prop["x0"] == priors["x0"].maximum + 0.5)
        self.assertEqual(prop["x0"], priors["x0"].minimum + 0.5)

    def test_periodic_boundary_below(self):
        priors = self.create_priors(boundary="periodic")
        chain = self.create_chain()
        proposal = GivenProposal(priors)

        sample = self.create_random_sample()
        sample["x0"] = priors["x0"].minimum - 0.5
        proposal.given_sample = copy.deepcopy(sample)

        prop, _ = proposal(chain)
        self.assertFalse(prop["x0"] == priors["x0"].minimum - 0.5)
        self.assertEqual(prop["x0"], priors["x0"].maximum - 0.5)


class TestProposals(TestBaseProposals):
    def setUp(self):
        self.outdir = "chain_test"
        if os.path.isdir(self.outdir) is False:
            os.mkdir(self.outdir)

    def tearDown(self):
        if os.path.isdir(self.outdir):
            shutil.rmtree(self.outdir)

    def get_simple_proposals(self):
        clsmembers = inspect.getmembers(
            sys.modules[proposals.__name__], inspect.isclass
        )
        clsmembers_clean = []
        for name, cls in clsmembers:
            a = "Proposal" in name
            b = "Base" not in name
            c = "Ensemble" not in name
            d = "Phase" not in name
            e = "Polarisation" not in name
            f = "Cycle" not in name
            g = "KDE" not in name
            h = "NormalizingFlow" not in name
            if a * b * c * d * e * f * g * h:
                clsmembers_clean.append((name, cls))

        return clsmembers_clean

    def proposal_check(self, prop, ndim=2, N=100):
        chain = self.create_chain(ndim=ndim)

        print(f"Testing {prop.__class__.__name__}")
        # Timing and return type
        start = time.time()
        for _ in range(N):
            p, w = prop(chain)
            chain.append(p)
        dt = 1e3 * (time.time() - start) / N
        print(f"Testing {prop.__class__.__name__}: dt~{dt:0.2g} [ms]")

        self.assertTrue(isinstance(p, Sample))
        self.assertTrue(isinstance(w, (int, float)))

    def test_proposal_return_type(self):
        priors = self.create_priors()
        for name, cls in self.get_simple_proposals():
            prop = cls(priors)
            self.proposal_check(prop)

    def test_KDE_proposal(self):
        priors = self.create_priors()
        prop = proposals.KDEProposal(priors)
        self.proposal_check(prop, N=20000)

    def test_GMM_proposal(self):
        if importlib.util.find_spec("sklearn") is not None:
            priors = self.create_priors()
            prop = proposals.GMMProposal(priors)
            self.proposal_check(prop, N=20000)
            self.assertTrue(prop.trained)
        else:
            print("Unable to test GMM as sklearn is not installed")

    def test_NF_proposal(self):
        priors = self.create_priors()
        chain = self.create_chain(10000)
        prop = proposals.NormalizingFlowProposal(priors, first_fit=10000)
        prop.steps_since_refit = 9999
        start = time.time()
        p, w = prop(chain)
        dt = time.time() - start
        print(f"Training for {prop.__class__.__name__} took dt~{dt:0.2g} [s]")
        self.assertTrue(prop.trained)

        self.proposal_check(prop)

    def test_NF_proposal_15D(self):
        ndim = 15
        priors = self.create_priors(ndim)
        chain = self.create_chain(10000, ndim=ndim)
        prop = proposals.NormalizingFlowProposal(priors, first_fit=10000)
        prop.steps_since_refit = 9999
        start = time.time()
        p, w = prop(chain)
        dt = time.time() - start
        print(f"Training for {prop.__class__.__name__} took dt~{dt:0.2g} [s]")
        self.assertTrue(prop.trained)

        self.proposal_check(prop, ndim=ndim)


if __name__ == "__main__":
    unittest.main()



import array
import copy
import sys

import numpy as np
from numpy.lib.recfunctions import structured_to_unstructured
from pandas import DataFrame

from ..utils import check_directory_exists_and_if_not_mkdir, logger
from .base_sampler import NestedSampler, signal_wrapper
from .proposal import JumpProposalCycle, Sample


class Cpnest(NestedSampler):
    """bilby wrapper of cpnest (https://github.com/johnveitch/cpnest)

    All positional and keyword arguments (i.e., the args and kwargs) passed to
    `run_sampler` will be propagated to `cpnest.CPNest`, see documentation
    for that class for further help. Under Other Parameters, we list commonly
    used kwargs and the bilby defaults.

    Parameters
    ==========
    nlive: int
        The number of live points, note this can also equivalently be given as
        one of [npoints, nlives, n_live_points]
    seed: int (1234)
        Initialised random seed
    nthreads: int, (1)
        Number of threads to use
    maxmcmc: int (1000)
        The maximum number of MCMC steps to take
    verbose: Bool (True)
        If true, print information information about the convergence during
    resume: Bool (True)
        Whether or not to resume from a previous run
    output: str
        Where to write the CPNest, by default this is
        {self.outdir}/cpnest_{self.label}/

    """

    default_kwargs = dict(
        verbose=3,
        nthreads=1,
        nlive=500,
        maxmcmc=1000,
        seed=None,
        poolsize=100,
        nhamiltonian=0,
        resume=True,
        output=None,
        proposals=None,
        n_periodic_checkpoint=8000,
    )
    hard_exit = True

    def _translate_kwargs(self, kwargs):
        if "nlive" not in kwargs:
            for equiv in self.npoints_equiv_kwargs:
                if equiv in kwargs:
                    kwargs["nlive"] = kwargs.pop(equiv)
        if "nthreads" not in kwargs:
            for equiv in self.npool_equiv_kwargs:
                if equiv in kwargs:
                    kwargs["nthreads"] = kwargs.pop(equiv)

        if "seed" not in kwargs:
            logger.warning("No seed provided, cpnest will use 1234.")

    @signal_wrapper
    def run_sampler(self):
        from cpnest import CPNest
        from cpnest import model as cpmodel
        from cpnest.nest2pos import compute_weights
        from cpnest.parameter import LivePoint

        class Model(cpmodel.Model):
            """A wrapper class to pass our log_likelihood into cpnest"""

            def __init__(self, names, priors):
                self.names = names
                self.priors = priors
                self._update_bounds()

            @staticmethod
            def log_likelihood(x, **kwargs):
                theta = [x[n] for n in self.search_parameter_keys]
                return self.log_likelihood(theta)

            @staticmethod
            def log_prior(x, **kwargs):
                theta = [x[n] for n in self.search_parameter_keys]
                return self.log_prior(theta)

            def _update_bounds(self):
                self.bounds = [
                    [self.priors[key].minimum, self.priors[key].maximum]
                    for key in self.names
                ]

            def new_point(self):
                """Draw a point from the prior"""
                prior_samples = self.priors.sample()
                self._update_bounds()
                point = LivePoint(
                    self.names,
                    array.array("d", [prior_samples[name] for name in self.names]),
                )
                return point

        self._resolve_proposal_functions()
        model = Model(self.search_parameter_keys, self.priors)
        out = None
        remove_kwargs = ["proposals", "n_periodic_checkpoint"]
        while out is None:
            try:
                out = CPNest(model, **self.kwargs)
            except TypeError as e:
                if len(remove_kwargs) > 0:
                    kwarg = remove_kwargs.pop(0)
                else:
                    raise TypeError("Unable to initialise cpnest sampler")
                logger.info(f"CPNest init. failed with error {e}, please update")
                logger.info(f"Attempting to rerun with kwarg {kwarg} removed")
                self.kwargs.pop(kwarg)
        try:
            out.run()
        except SystemExit:
            out.checkpoint()
            self.write_current_state_and_exit()

        if self.plot:
            out.plot()

        self.calc_likelihood_count()
        self.result.samples = structured_to_unstructured(
            out.posterior_samples[self.search_parameter_keys]
        )
        self.result.log_likelihood_evaluations = out.posterior_samples["logL"]
        self.result.nested_samples = DataFrame(out.get_nested_samples(filename=""))
        self.result.nested_samples.rename(
            columns=dict(logL="log_likelihood"), inplace=True
        )
        _, log_weights = compute_weights(
            np.array(self.result.nested_samples.log_likelihood),
            np.array(out.NS.state.nlive),
        )
        self.result.nested_samples["weights"] = np.exp(log_weights)
        self.result.log_evidence = out.NS.state.logZ
        self.result.log_evidence_err = np.sqrt(out.NS.state.info / out.NS.state.nlive)
        self.result.information_gain = out.NS.state.info
        return self.result

    def write_current_state_and_exit(self, signum=None, frame=None):
        """
        Overwrites the base class to make sure that :code:`CPNest` terminates
        properly as :code:`CPNest` handles all the multiprocessing internally.
        """
        self._log_interruption(signum=signum)
        sys.exit(self.exit_code)

    def _verify_kwargs_against_default_kwargs(self):
        """
        Set the directory where the output will be written
        and check resume and checkpoint status.
        """
        if not self.kwargs["output"]:
            self.kwargs["output"] = f"{self.outdir}/cpnest_{self.label}/"
        if self.kwargs["output"].endswith("/") is False:
            self.kwargs["output"] = f"{self.kwargs['output']}/"
        check_directory_exists_and_if_not_mkdir(self.kwargs["output"])
        if self.kwargs["n_periodic_checkpoint"] and not self.kwargs["resume"]:
            self.kwargs["n_periodic_checkpoint"] = None
        NestedSampler._verify_kwargs_against_default_kwargs(self)

    def _resolve_proposal_functions(self):
        from cpnest.proposal import ProposalCycle

        if "proposals" in self.kwargs:
            if self.kwargs["proposals"] is None:
                return
            if type(self.kwargs["proposals"]) == JumpProposalCycle:
                self.kwargs["proposals"] = dict(
                    mhs=self.kwargs["proposals"], hmc=self.kwargs["proposals"]
                )
            for key, proposal in self.kwargs["proposals"].items():
                if isinstance(proposal, JumpProposalCycle):
                    self.kwargs["proposals"][key] = cpnest_proposal_cycle_factory(
                        proposal
                    )
                elif isinstance(proposal, ProposalCycle):
                    pass
                else:
                    raise TypeError("Unknown proposal type")


def cpnest_proposal_factory(jump_proposal):
    import cpnest.proposal

    class CPNestEnsembleProposal(cpnest.proposal.EnsembleProposal):
        def __init__(self, jp):
            self.jump_proposal = jp
            self.ensemble = None

        def __call__(self, sample, **kwargs):
            return self.get_sample(sample, **kwargs)

        def get_sample(self, cpnest_sample, **kwargs):
            sample = Sample.from_cpnest_live_point(cpnest_sample)
            self.ensemble = kwargs.get("coordinates", self.ensemble)
            sample = self.jump_proposal(sample=sample, sampler_name="cpnest", **kwargs)
            self.log_J = self.jump_proposal.log_j
            return self._update_cpnest_sample(cpnest_sample, sample)

        @staticmethod
        def _update_cpnest_sample(cpnest_sample, sample):
            cpnest_sample.names = list(sample.keys())
            for i, value in enumerate(sample.values()):
                cpnest_sample.values[i] = value
            return cpnest_sample

    return CPNestEnsembleProposal(jump_proposal)


def cpnest_proposal_cycle_factory(jump_proposals):
    import cpnest.proposal

    class CPNestProposalCycle(cpnest.proposal.ProposalCycle):
        def __init__(self):
            self.jump_proposals = copy.deepcopy(jump_proposals)
            for i, prop in enumerate(self.jump_proposals.proposal_functions):
                self.jump_proposals.proposal_functions[i] = cpnest_proposal_factory(
                    prop
                )
            self.jump_proposals.update_cycle()
            super(CPNestProposalCycle, self).__init__(
                proposals=self.jump_proposals.proposal_functions,
                weights=self.jump_proposals.weights,
                cyclelength=self.jump_proposals.cycle_length,
            )

        def get_sample(self, old, **kwargs):
            return self.jump_proposals(sample=old, coordinates=self.ensemble, **kwargs)

        def set_ensemble(self, ensemble):
            self.ensemble = ensemble

    return CPNestProposalCycle



import argparse
import logging
import os
import subprocess

from .log import logger


def set_up_command_line_arguments():
    """ Sets up command line arguments that can be used to modify how scripts are run.

    Returns
    =======
    command_line_args, command_line_parser: tuple
        The command_line_args is a Namespace of the command line arguments while
        the command_line_parser can be given to a new `argparse.ArgumentParser`
        as a parent object from which to inherit.

    Notes
    =====
        The command line arguments are passed initially at runtime, but this parser
        does not have a `--help` option (i.e., the command line options are
        available for any script which includes `import bilby`, but no help command
        is available. This is done to avoid conflicts with child argparse routines
        (see the example below).

    Examples
    ========
    In the following example we demonstrate how to setup a custom command line for a
    project which uses bilby.

    .. code-block:: python

        # Here we import bilby, which initialises and parses the default command-line args
        >>> import bilby
        # The command line arguments can then be accessed via
        >>> bilby.core.utils.command_line_args
        Namespace(clean=False, log_level=20, quite=False)
        # Next, we import argparse and define a new argparse object
        >>> import argparse
        >>> parser = argparse.ArgumentParser(parents=[bilby.core.utils.command_line_parser])
        >>> parser.add_argument('--argument', type=int, default=1)
        >>> args = parser.parse_args()
        Namespace(clean=False, log_level=20, quite=False, argument=1)

    Placing these lines into a script, you'll be able to pass in the usual bilby default
    arguments, in addition to `--argument`. To see a list of all options, call the script
    with `--help`.

    """
    try:
        parser = argparse.ArgumentParser(
            description="Command line interface for bilby scripts",
            add_help=False, allow_abbrev=False)
    except TypeError:
        parser = argparse.ArgumentParser(
            description="Command line interface for bilby scripts",
            add_help=False)
    parser.add_argument("-v", "--verbose", action="store_true",
                        help=("Increase output verbosity [logging.DEBUG]." +
                              " Overridden by script level settings"))
    parser.add_argument("-q", "--quiet", action="store_true",
                        help=("Decrease output verbosity [logging.WARNING]." +
                              " Overridden by script level settings"))
    parser.add_argument("-c", "--clean", action="store_true",
                        help="Force clean data, never use cached data")
    parser.add_argument("-u", "--use-cached", action="store_true",
                        help="Force cached data and do not check its validity")
    parser.add_argument("--sampler-help", nargs='?', default=False,
                        const='None', help="Print help for given sampler")
    parser.add_argument("--bilby-test-mode", action="store_true",
                        help=("Used for testing only: don't run full PE, but"
                              " just check nothing breaks"))
    parser.add_argument("--bilby-zero-likelihood-mode", action="store_true",
                        help=("Used for testing only: don't run full PE, but"
                              " just check nothing breaks"))
    args, unknown_args = parser.parse_known_args()
    if args.quiet:
        args.log_level = logging.WARNING
    elif args.verbose:
        args.log_level = logging.DEBUG
    else:
        args.log_level = logging.INFO

    return args, parser


def run_commandline(cl, log_level=20, raise_error=True, return_output=True):
    """Run a string cmd as a subprocess, check for errors and return output.

    Parameters
    ==========
    cl: str
        Command to run
    log_level: int
        See https://docs.python.org/2/library/logging.html#logging-levels,
        default is '20' (INFO)

    """

    logger.log(log_level, 'Now executing: ' + cl)
    if return_output:
        try:
            out = subprocess.check_output(
                cl, stderr=subprocess.STDOUT, shell=True,
                universal_newlines=True)
        except subprocess.CalledProcessError as e:
            logger.log(log_level, 'Execution failed: {}'.format(e.output))
            if raise_error:
                raise
            else:
                out = 0
        os.system('\n')
        return out
    else:
        process = subprocess.Popen(cl, shell=True)
        process.communicate()



import os
import shutil
from shutil import copyfile

import numpy as np

from .base_sampler import SamplerError, signal_wrapper
from .emcee import Emcee
from .ptemcee import LikePriorEvaluator

_evaluator = LikePriorEvaluator()


class Zeus(Emcee):
    """bilby wrapper for Zeus (https://zeus-mcmc.readthedocs.io/)

    All positional and keyword arguments (i.e., the args and kwargs) passed to
    `run_sampler` will be propagated to `zeus.EnsembleSampler`, see
    documentation for that class for further help. Under Other Parameters, we
    list commonly used kwargs and the bilby defaults.

    Parameters
    ==========
    nwalkers: int, (500)
        The number of walkers
    nsteps: int, (100)
        The number of steps
    nburn: int (None)
        If given, the fixed number of steps to discard as burn-in. These will
        be discarded from the total number of steps set by `nsteps` and
        therefore the value must be greater than `nsteps`. Else, nburn is
        estimated from the autocorrelation time
    burn_in_fraction: float, (0.25)
        The fraction of steps to discard as burn-in in the event that the
        autocorrelation time cannot be calculated
    burn_in_act: float
        The number of autocorrelation times to discard as burn-in

    """

    default_kwargs = dict(
        nwalkers=500,
        args=[],
        kwargs={},
        pool=None,
        log_prob0=None,
        start=None,
        blobs0=None,
        iterations=100,
        thin=1,
    )

    def __init__(
        self,
        likelihood,
        priors,
        outdir="outdir",
        label="label",
        use_ratio=False,
        plot=False,
        skip_import_verification=False,
        pos0=None,
        nburn=None,
        burn_in_fraction=0.25,
        resume=True,
        burn_in_act=3,
        **kwargs,
    ):
        super(Zeus, self).__init__(
            likelihood=likelihood,
            priors=priors,
            outdir=outdir,
            label=label,
            use_ratio=use_ratio,
            plot=plot,
            skip_import_verification=skip_import_verification,
            pos0=pos0,
            nburn=nburn,
            burn_in_fraction=burn_in_fraction,
            resume=resume,
            burn_in_act=burn_in_act,
            **kwargs,
        )

    def _translate_kwargs(self, kwargs):
        super(Zeus, self)._translate_kwargs(kwargs=kwargs)

        # check if using emcee-style arguments
        if "start" not in kwargs:
            if "rstate0" in kwargs:
                kwargs["start"] = kwargs.pop("rstate0")
        if "log_prob0" not in kwargs:
            if "lnprob0" in kwargs:
                kwargs["log_prob0"] = kwargs.pop("lnprob0")

    @property
    def sampler_function_kwargs(self):
        keys = ["log_prob0", "start", "blobs0", "iterations", "thin", "progress"]

        function_kwargs = {key: self.kwargs[key] for key in keys if key in self.kwargs}

        return function_kwargs

    @property
    def sampler_init_kwargs(self):
        init_kwargs = {
            key: value
            for key, value in self.kwargs.items()
            if key not in self.sampler_function_kwargs
        }

        init_kwargs["logprob_fn"] = _evaluator.call_emcee
        init_kwargs["ndim"] = self.ndim

        return init_kwargs

    def write_current_state(self):
        self._sampler.distribute = map
        super(Zeus, self).write_current_state()
        self._sampler.distribute = getattr(self._sampler.pool, "map", map)

    def _initialise_sampler(self):
        from zeus import EnsembleSampler

        self._sampler = EnsembleSampler(**self.sampler_init_kwargs)
        self._init_chain_file()

    def write_chains_to_file(self, sample):
        chain_file = self.checkpoint_info.chain_file
        temp_chain_file = chain_file + ".temp"
        if os.path.isfile(chain_file):
            copyfile(chain_file, temp_chain_file)

        points = np.hstack([sample[0], np.array(sample[2])])

        with open(temp_chain_file, "a") as ff:
            for ii, point in enumerate(points):
                ff.write(self.checkpoint_info.chain_template.format(ii, *point))
        shutil.move(temp_chain_file, chain_file)

    def _set_pos0_for_resume(self):
        self.pos0 = self.sampler.get_last_sample()

    @signal_wrapper
    def run_sampler(self):
        self._setup_pool()
        sampler_function_kwargs = self.sampler_function_kwargs
        iterations = sampler_function_kwargs.pop("iterations")
        iterations -= self._previous_iterations

        sampler_function_kwargs["start"] = self.pos0

        # main iteration loop
        for sample in self.sampler.sample(
            iterations=iterations, **sampler_function_kwargs
        ):
            self.write_chains_to_file(sample)
        self._close_pool()
        self.write_current_state()

        self.result.sampler_output = np.nan
        self.calculate_autocorrelation(self.sampler.chain.reshape((-1, self.ndim)))
        self.print_nburn_logging_info()

        self._generate_result()

        self.result.samples = self.sampler.get_chain(flat=True, discard=self.nburn)
        self.result.walkers = self.sampler.chain
        return self.result

    def _generate_result(self):
        self.result.nburn = self.nburn
        self.calc_likelihood_count()
        if self.result.nburn > self.nsteps:
            raise SamplerError(
                "The run has finished, but the chain is not burned in: "
                f"`nburn < nsteps` ({self.result.nburn} < {self.nsteps})."
                " Try increasing the number of steps."
            )
        blobs = np.array(self.sampler.get_blobs(flat=True, discard=self.nburn)).reshape(
            (-1, 2)
        )
        log_likelihoods, log_priors = blobs.T
        self.result.log_likelihood_evaluations = log_likelihoods
        self.result.log_prior_evaluations = log_priors
        self.result.log_evidence = np.nan
        self.result.log_evidence_err = np.nan



import importlib
import time
from abc import ABCMeta, abstractmethod

import numpy as np
from scipy.spatial.distance import jensenshannon
from scipy.stats import gaussian_kde

from ..core.prior import PriorDict
from ..core.sampler.base_sampler import SamplerError
from ..core.utils import logger, reflect
from ..gw.source import PARAMETER_SETS


class ProposalCycle(object):
    def __init__(self, proposal_list):
        self.proposal_list = proposal_list
        self.weights = [prop.weight for prop in self.proposal_list]
        self.normalized_weights = [w / sum(self.weights) for w in self.weights]
        self.weighted_proposal_list = [
            np.random.choice(self.proposal_list, p=self.normalized_weights)
            for _ in range(10 * int(1 / min(self.normalized_weights)))
        ]
        self.nproposals = len(self.weighted_proposal_list)
        self._position = 0

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, position):
        self._position = np.mod(position, self.nproposals)

    def get_proposal(self):
        prop = self.weighted_proposal_list[self._position]
        self.position += 1
        return prop

    def __str__(self):
        string = "ProposalCycle:\n"
        for prop in self.proposal_list:
            string += f"  {prop}\n"
        return string


class BaseProposal(object):
    _accepted = 0
    _rejected = 0
    __metaclass__ = ABCMeta

    def __init__(self, priors, weight=1, subset=None):
        self._str_attrs = ["acceptance_ratio", "n"]

        self.parameters = priors.non_fixed_keys
        self.weight = weight
        self.subset = subset

        # Restrict to a subset
        if self.subset is not None:
            self.parameters = [p for p in self.parameters if p in subset]
            self._str_attrs.append("parameters")

        self.ndim = len(self.parameters)

        self.prior_boundary_dict = {key: priors[key].boundary for key in priors}
        self.prior_minimum_dict = {key: np.max(priors[key].minimum) for key in priors}
        self.prior_maximum_dict = {key: np.min(priors[key].maximum) for key in priors}
        self.prior_width_dict = {key: np.max(priors[key].width) for key in priors}

    @property
    def accepted(self):
        return self._accepted

    @accepted.setter
    def accepted(self, accepted):
        self._accepted = accepted

    @property
    def rejected(self):
        return self._rejected

    @rejected.setter
    def rejected(self, rejected):
        self._rejected = rejected

    @property
    def acceptance_ratio(self):
        if self.n == 0:
            return np.nan
        else:
            return self.accepted / self.n

    @property
    def n(self):
        return self.accepted + self.rejected

    def __str__(self):
        msg = [f"{type(self).__name__}("]
        for attr in self._str_attrs:
            val = getattr(self, attr, "N/A")
            if isinstance(val, (float, int)):
                val = f"{val:1.2g}"
            msg.append(f"{attr}:{val},")
        return "".join(msg) + ")"

    def apply_boundaries(self, point):
        for key in self.parameters:
            boundary = self.prior_boundary_dict[key]
            if boundary is None:
                continue
            elif boundary == "periodic":
                point[key] = self.apply_periodic_boundary(key, point[key])
            elif boundary == "reflective":
                point[key] = self.apply_reflective_boundary(key, point[key])
            else:
                raise SamplerError(f"Boundary {boundary} not implemented")
        return point

    def apply_periodic_boundary(self, key, val):
        minimum = self.prior_minimum_dict[key]
        width = self.prior_width_dict[key]
        return minimum + np.mod(val - minimum, width)

    def apply_reflective_boundary(self, key, val):
        minimum = self.prior_minimum_dict[key]
        width = self.prior_width_dict[key]
        val_normalised = (val - minimum) / width
        val_normalised_reflected = reflect(np.array(val_normalised))
        return minimum + width * val_normalised_reflected

    def __call__(self, chain):
        sample, log_factor = self.propose(chain)
        sample = self.apply_boundaries(sample)
        return sample, log_factor

    @abstractmethod
    def propose(self, chain):
        """Propose a new point

        This method must be overwritten by implemented proposals. The propose
        method is called by __call__, then boundaries applied, before returning
        the proposed point.

        Parameters
        ----------
        chain: bilby.core.sampler.bilby_mcmc.chain.Chain
            The chain to use for the proposal

        Returns
        -------
        proposal: bilby.core.sampler.bilby_mcmc.Sample
            The proposed point
        log_factor: float
            The natural-log of the additional factor entering the acceptance
            probability to ensure detailed balance. For symmetric proposals,
            a value of 0 should be returned.
        """
        pass

    @staticmethod
    def check_dependencies(warn=True):
        """Check the dependencies required to use the proposal

        Parameters
        ----------
        warn: bool
            If true, print a warning

        Returns
        -------
        check: bool
            If true, dependencies exist
        """
        return True


class FixedGaussianProposal(BaseProposal):
    """A proposal using a fixed non-correlated Gaussian distribution

    Parameters
    ----------
    priors: bilby.core.prior.PriorDict
        The set of priors
    weight: float
        Weighting factor
    subset: list
        A list of keys for which to restrict the proposal to (other parameters
        will be kept fixed)
    sigma: float
        The scaling factor for proposals
    """

    def __init__(self, priors, weight=1, subset=None, sigma=0.01):
        super(FixedGaussianProposal, self).__init__(priors, weight, subset)
        self.sigmas = {}
        for key in self.parameters:
            if np.isinf(self.prior_width_dict[key]):
                self.prior_width_dict[key] = 1
            if isinstance(sigma, float):
                self.sigmas[key] = sigma
            elif isinstance(sigma, dict):
                self.sigmas[key] = sigma[key]
            else:
                raise SamplerError("FixedGaussianProposal sigma not understood")

    def propose(self, chain):
        sample = chain.current_sample
        for key in self.parameters:
            sigma = self.prior_width_dict[key] * self.sigmas[key]
            sample[key] += sigma * np.random.randn()
        log_factor = 0
        return sample, log_factor


class AdaptiveGaussianProposal(BaseProposal):
    def __init__(
        self,
        priors,
        weight=1,
        subset=None,
        sigma=1,
        scale_init=1e0,
        stop=1e5,
        target_facc=0.234,
    ):
        super(AdaptiveGaussianProposal, self).__init__(priors, weight, subset)
        self.sigmas = {}
        for key in self.parameters:
            if np.isinf(self.prior_width_dict[key]):
                self.prior_width_dict[key] = 1
            if isinstance(sigma, (float, int)):
                self.sigmas[key] = sigma
            elif isinstance(sigma, dict):
                self.sigmas[key] = sigma[key]
            else:
                raise SamplerError("AdaptiveGaussianProposal sigma not understood")

        self.target_facc = target_facc
        self.scale = scale_init
        self.stop = stop
        self._str_attrs.append("scale")
        self._last_accepted = 0

    def propose(self, chain):
        sample = chain.current_sample
        self.update_scale(chain)
        if np.random.random() < 1e-3:
            factor = 1e1
        elif np.random.random() < 1e-4:
            factor = 1e2
        else:
            factor = 1
        for key in self.parameters:
            sigma = factor * self.scale * self.prior_width_dict[key] * self.sigmas[key]
            sample[key] += sigma * np.random.randn()
        log_factor = 0
        return sample, log_factor

    def update_scale(self, chain):
        """
        The adaptation of the scale follows (35)/(36) of https://arxiv.org/abs/1409.7215
        """
        if 0 < self.n < self.stop:
            s_gamma = (self.stop / self.n) ** 0.2 - 1
            if self.accepted > self._last_accepted:
                self.scale += s_gamma * (1 - self.target_facc) / 100
            else:
                self.scale -= s_gamma * self.target_facc / 100
            self._last_accepted = self.accepted
            self.scale = max(self.scale, 1 / self.stop)


class DifferentialEvolutionProposal(BaseProposal):
    """A proposal using Differential Evolution

    Parameters
    ----------
    priors: bilby.core.prior.PriorDict
        The set of priors
    weight: float
        Weighting factor
    subset: list
        A list of keys for which to restrict the proposal to (other parameters
        will be kept fixed)
    mode_hopping_frac: float
        The fraction of proposals which use 'mode hopping'
    """

    def __init__(self, priors, weight=1, subset=None, mode_hopping_frac=0.5):
        super(DifferentialEvolutionProposal, self).__init__(priors, weight, subset)
        self.mode_hopping_frac = mode_hopping_frac

    def propose(self, chain):
        theta = chain.current_sample
        theta1 = chain.random_sample
        theta2 = chain.random_sample
        if np.random.rand() > self.mode_hopping_frac:
            gamma = 1
        else:
            # Base jump size
            gamma = np.random.normal(0, 2.38 / np.sqrt(2 * self.ndim))
            # Scale uniformly in log between 0.1 and 10 times
            gamma *= np.exp(np.log(0.1) + np.log(100.0) * np.random.rand())

        for key in self.parameters:
            theta[key] += gamma * (theta2[key] - theta1[key])

        log_factor = 0
        return theta, log_factor


class UniformProposal(BaseProposal):
    """A proposal using uniform draws from the prior support

    Note: for priors with infinite support, this proposal will not propose a
    point, leading to inefficient sampling. You may wish to omit this proposal
    if you have priors with infinite support.

    Parameters
    ----------
    priors: bilby.core.prior.PriorDict
        The set of priors
    weight: float
        Weighting factor
    subset: list
        A list of keys for which to restrict the proposal to (other parameters
        will be kept fixed)
    """

    def __init__(self, priors, weight=1, subset=None):
        super(UniformProposal, self).__init__(priors, weight, subset)

    def propose(self, chain):
        sample = chain.current_sample
        for key in self.parameters:
            width = self.prior_width_dict[key]
            if np.isinf(width) is False:
                sample[key] = np.random.uniform(
                    self.prior_minimum_dict[key], self.prior_maximum_dict[key]
                )
            else:
                # Unable to generate a uniform sample on infinite support
                pass
        log_factor = 0
        return sample, log_factor


class PriorProposal(BaseProposal):
    """A proposal using draws from the prior distribution

    Note: for priors which use interpolation, this proposal can be problematic
    as the proposal gets pickled in multiprocessing. Either, use serial
    processing (npool=1) or fall back to a UniformProposal.

    Parameters
    ----------
    priors: bilby.core.prior.PriorDict
        The set of priors
    weight: float
        Weighting factor
    subset: list
        A list of keys for which to restrict the proposal to (other parameters
        will be kept fixed)
    """

    def __init__(self, priors, weight=1, subset=None):
        super(PriorProposal, self).__init__(priors, weight, subset)
        self.priors = PriorDict({key: priors[key] for key in self.parameters})

    def propose(self, chain):
        sample = chain.current_sample
        lnp_theta = self.priors.ln_prob(sample.as_dict(self.parameters))
        prior_sample = self.priors.sample()
        for key in self.parameters:
            sample[key] = prior_sample[key]
        lnp_thetaprime = self.priors.ln_prob(sample.as_dict(self.parameters))
        log_factor = lnp_theta - lnp_thetaprime
        return sample, log_factor


_density_estimate_doc = """ A proposal using draws from a {estimator} fit to the chain

Parameters
----------
priors: bilby.core.prior.PriorDict
    The set of priors
weight: float
    Weighting factor
subset: list
    A list of keys for which to restrict the proposal to (other parameters
    will be kept fixed)
first_fit: int
    The number of steps to take before first fitting the KDE
fit_multiplier: int
    The multiplier for the next fit
nsamples_for_density: int
    The number of samples to use when fitting the KDE
fallback: bilby.core.sampler.bilby_mcmc.proposal.BaseProposal
    A proposal to use before first training
scale_fits: int
    A scaling factor for both the initial and subsequent updates
"""


class DensityEstimateProposal(BaseProposal):
    def __init__(
        self,
        priors,
        weight=1,
        subset=None,
        first_fit=1000,
        fit_multiplier=10,
        nsamples_for_density=1000,
        fallback=AdaptiveGaussianProposal,
        scale_fits=1,
    ):
        super(DensityEstimateProposal, self).__init__(priors, weight, subset)
        self.nsamples_for_density = nsamples_for_density
        self.fallback = fallback(priors, weight, subset)
        self.fit_multiplier = fit_multiplier * scale_fits

        # Counters
        self.steps_since_refit = 0
        self.next_refit_time = first_fit * scale_fits
        self.density = None
        self.trained = False
        self._str_attrs.append("trained")

    density_name = None
    __doc__ = _density_estimate_doc.format(estimator=density_name)

    def _fit(self, dataset):
        raise NotImplementedError

    def _evaluate(self, point):
        raise NotImplementedError

    def _sample(self, nsamples=None):
        raise NotImplementedError

    def refit(self, chain):
        current_density = self.density
        start = time.time()

        # Draw two (possibly overlapping) data sets for training and verification
        dataset = []
        verification_dataset = []
        nsamples_for_density = min(chain.position, self.nsamples_for_density)
        for _ in range(nsamples_for_density):
            s = chain.random_sample
            dataset.append([s[key] for key in self.parameters])
            s = chain.random_sample
            verification_dataset.append([s[key] for key in self.parameters])

        # Fit the density
        self.density = self._fit(np.array(dataset).T)

        # Print a log message
        took = time.time() - start
        logger.info(
            f"{self.density_name} construction at {self.steps_since_refit} finished"
            f" for length {chain.position} chain, took {took:0.2f}s."
            f" Current accept-ratio={self.acceptance_ratio:0.2f}"
        )

        # Reset counters for next training
        self.steps_since_refit = 0
        self.next_refit_time *= self.fit_multiplier

        # Verify training hasn't overconstrained
        new_draws = np.atleast_2d(self._sample(1000))
        verification_dataset = np.array(verification_dataset)
        fail_parameters = []
        for ii, key in enumerate(self.parameters):
            std_draws = np.std(new_draws[:, ii])
            std_verification = np.std(verification_dataset[:, ii])
            if std_draws < 0.1 * std_verification:
                fail_parameters.append(key)

        if len(fail_parameters) > 0:
            logger.info(
                f"{self.density_name} construction failed verification and is discarded"
            )
            self.density = current_density
        else:
            self.trained = True

    def propose(self, chain):
        self.steps_since_refit += 1

        # Check if we refit
        testA = self.steps_since_refit >= self.next_refit_time
        if testA:
            self.refit(chain)

        # If KDE is yet to be fitted, use the fallback
        if self.trained is False:
            return self.fallback.propose(chain)

        # Grab the current sample and it's probability under the KDE
        theta = chain.current_sample
        ln_p_theta = self._evaluate(list(theta.as_dict(self.parameters).values()))

        # Sample and update theta
        new_sample = self._sample(1)
        for key, val in zip(self.parameters, new_sample):
            theta[key] = val

        # Calculate the probability of the new sample and the KDE
        ln_p_thetaprime = self._evaluate(list(theta.as_dict(self.parameters).values()))

        # Calculate Q(theta|theta') / Q(theta'|theta)
        log_factor = ln_p_theta - ln_p_thetaprime

        return theta, log_factor


class KDEProposal(DensityEstimateProposal):
    density_name = "Gaussian KDE"
    __doc__ = _density_estimate_doc.format(estimator=density_name)

    def _fit(self, dataset):
        return gaussian_kde(dataset)

    def _evaluate(self, point):
        return self.density.logpdf(point)[0]

    def _sample(self, nsamples=None):
        return np.atleast_1d(np.squeeze(self.density.resample(nsamples)))


class GMMProposal(DensityEstimateProposal):
    density_name = "Gaussian Mixture Model"
    __doc__ = _density_estimate_doc.format(estimator=density_name)

    def _fit(self, dataset):
        from sklearn.mixture import GaussianMixture

        density = GaussianMixture(n_components=10)
        density.fit(dataset.T)
        return density

    def _evaluate(self, point):
        return np.squeeze(self.density.score_samples(np.atleast_2d(point)))

    def _sample(self, nsamples=None):
        return np.squeeze(self.density.sample(n_samples=nsamples)[0])

    def check_dependencies(warn=True):
        if importlib.util.find_spec("sklearn") is None:
            if warn:
                logger.warning(
                    "Unable to utilise GMMProposal as sklearn is not installed"
                )
            return False
        else:
            return True


class NormalizingFlowProposal(DensityEstimateProposal):
    density_name = "Normalizing Flow"
    __doc__ = _density_estimate_doc.format(estimator=density_name) + (
        """
        js_factor: float
            The factor to use in determining the max-JS factor to terminate
            training.
        max_training_epochs: int
            The maximum bumber of traning steps to take
        """
    )

    def __init__(
        self,
        priors,
        weight=1,
        subset=None,
        first_fit=1000,
        fit_multiplier=10,
        max_training_epochs=1000,
        scale_fits=1,
        nsamples_for_density=1000,
        js_factor=10,
        fallback=AdaptiveGaussianProposal,
    ):
        super(NormalizingFlowProposal, self).__init__(
            priors=priors,
            weight=weight,
            subset=subset,
            first_fit=first_fit,
            fit_multiplier=fit_multiplier,
            nsamples_for_density=nsamples_for_density,
            fallback=fallback,
            scale_fits=scale_fits,
        )
        self.setup_flow()
        self.setup_optimizer()

        self.max_training_epochs = max_training_epochs
        self.js_factor = js_factor

    def setup_flow(self):
        if self.ndim < 3:
            self.setup_basic_flow()
        else:
            self.setup_NVP_flow()

    def setup_NVP_flow(self):
        from .flows import NVPFlow

        self.flow = NVPFlow(
            features=self.ndim,
            hidden_features=self.ndim * 2,
            num_layers=2,
            num_blocks_per_layer=2,
            batch_norm_between_layers=True,
            batch_norm_within_layers=True,
        )

    def setup_basic_flow(self):
        from .flows import BasicFlow

        self.flow = BasicFlow(features=self.ndim)

    def setup_optimizer(self):
        from torch import optim

        self.optimizer = optim.Adam(self.flow.parameters())

    def get_training_data(self, chain):
        training_data = []
        nsamples_for_density = min(chain.position, self.nsamples_for_density)
        for _ in range(nsamples_for_density):
            s = chain.random_sample
            training_data.append([s[key] for key in self.parameters])
        return training_data

    def _calculate_js(self, validation_samples, training_samples_draw):
        # Calculate the maximum JS between the validation and draw
        max_js = 0
        for i in range(self.ndim):
            A = validation_samples[:, i]
            B = training_samples_draw[:, i]
            xmin = np.min([np.min(A), np.min(B)])
            xmax = np.min([np.max(A), np.max(B)])
            xval = np.linspace(xmin, xmax, 100)
            Apdf = gaussian_kde(A)(xval)
            Bpdf = gaussian_kde(B)(xval)
            js = jensenshannon(Apdf, Bpdf)
            max_js = max(max_js, js)
        return np.power(max_js, 2)

    def train(self, chain):
        logger.info("Starting NF training")

        import torch

        start = time.time()

        training_samples = np.array(self.get_training_data(chain))
        validation_samples = np.array(self.get_training_data(chain))

        training_tensor = torch.tensor(training_samples, dtype=torch.float32)

        max_js_threshold = self.js_factor / self.nsamples_for_density

        for epoch in range(1, self.max_training_epochs + 1):
            self.optimizer.zero_grad()
            loss = -self.flow.log_prob(inputs=training_tensor).mean()
            loss.backward()
            self.optimizer.step()

            # Draw from the current flow
            self.flow.eval()
            training_samples_draw = (
                self.flow.sample(self.nsamples_for_density).detach().numpy()
            )
            self.flow.train()

            if np.mod(epoch, 10) == 0:
                max_js_bits = self._calculate_js(
                    validation_samples, training_samples_draw
                )
                if max_js_bits < max_js_threshold:
                    logger.info(
                        f"Training complete after {epoch} steps, "
                        f"max_js_bits={max_js_bits:0.5f}<{max_js_threshold}"
                    )
                    break

        took = time.time() - start
        logger.info(
            f"Flow training step ({self.steps_since_refit}) finished"
            f" for length {chain.position} chain, took {took:0.2f}s."
            f" Current accept-ratio={self.acceptance_ratio:0.2f}"
        )
        self.steps_since_refit = 0
        self.next_refit_time *= self.fit_multiplier
        self.trained = True

    def propose(self, chain):
        import torch

        self.steps_since_refit += 1
        theta = chain.current_sample

        # Check if we retrain the NF
        testA = self.steps_since_refit >= self.next_refit_time
        if testA:
            self.train(chain)

        if self.trained is False:
            return self.fallback.propose(chain)

        self.flow.eval()
        theta_prime_T = self.flow.sample(1)

        logp_theta_prime = self.flow.log_prob(theta_prime_T).detach().numpy()[0]
        theta_T = torch.tensor(
            np.atleast_2d([theta[key] for key in self.parameters]), dtype=torch.float32
        )
        logp_theta = self.flow.log_prob(theta_T).detach().numpy()[0]
        log_factor = logp_theta - logp_theta_prime

        flow_sample_values = np.atleast_1d(np.squeeze(theta_prime_T.detach().numpy()))
        for key, val in zip(self.parameters, flow_sample_values):
            theta[key] = val

        return theta, float(log_factor)

    def check_dependencies(warn=True):
        if importlib.util.find_spec("nflows") is None:
            if warn:
                logger.warning(
                    "Unable to utilise NormalizingFlowProposal as nflows is not installed"
                )
            return False
        else:
            return True


class FixedJumpProposal(BaseProposal):
    def __init__(self, priors, jumps=1, subset=None, weight=1, scale=1e-4):
        super(FixedJumpProposal, self).__init__(priors, weight, subset)
        self.scale = scale
        if isinstance(jumps, (int, float)):
            self.jumps = {key: jumps for key in self.parameters}
        elif isinstance(jumps, dict):
            self.jumps = jumps
        else:
            raise SamplerError("jumps not understood")

    def propose(self, chain):
        sample = chain.current_sample
        for key, jump in self.jumps.items():
            sign = np.random.randint(2) * 2 - 1
            sample[key] += sign * jump + self.epsilon * self.prior_width_dict[key]
        log_factor = 0
        return sample, log_factor

    @property
    def epsilon(self):
        return self.scale * np.random.normal()


class BaseGravitationalWaveTransientProposal(BaseProposal):
    def __init__(self, priors, weight=1):
        super(BaseGravitationalWaveTransientProposal, self).__init__(
            priors, weight=weight
        )
        if "phase" in priors:
            self.phase_key = "phase"
        elif "delta_phase" in priors:
            self.phase_key = "delta_phase"
        else:
            self.phase_key = None

    def get_cos_theta_jn(self, sample):
        if "cos_theta_jn" in sample.parameter_keys:
            cos_theta_jn = sample["cos_theta_jn"]
        elif "theta_jn" in sample.parameter_keys:
            cos_theta_jn = np.cos(sample["theta_jn"])
        else:
            raise SamplerError()
        return cos_theta_jn

    def get_phase(self, sample):
        if "phase" in sample.parameter_keys:
            return sample["phase"]
        elif "delta_phase" in sample.parameter_keys:
            cos_theta_jn = self.get_cos_theta_jn(sample)
            delta_phase = sample["delta_phase"]
            psi = sample["psi"]
            phase = np.mod(delta_phase - np.sign(cos_theta_jn) * psi, 2 * np.pi)
        else:
            raise SamplerError()
        return phase

    def get_delta_phase(self, phase, sample):
        cos_theta_jn = self.get_cos_theta_jn(sample)
        psi = sample["psi"]
        delta_phase = phase + np.sign(cos_theta_jn) * psi
        return delta_phase


class CorrelatedPolarisationPhaseJump(BaseGravitationalWaveTransientProposal):
    def __init__(self, priors, weight=1):
        super(CorrelatedPolarisationPhaseJump, self).__init__(priors, weight=weight)

    def propose(self, chain):
        sample = chain.current_sample
        phase = self.get_phase(sample)

        alpha = sample["psi"] + phase
        beta = sample["psi"] - phase

        draw = np.random.random()
        if draw < 0.5:
            alpha = 3.0 * np.pi * np.random.random()
        else:
            beta = 3.0 * np.pi * np.random.random() - 2 * np.pi

        # Update
        sample["psi"] = (alpha + beta) * 0.5
        phase = (alpha - beta) * 0.5

        if self.phase_key == "delta_phase":
            sample["delta_phase"] = self.get_delta_phase(phase, sample)
        else:
            sample["phase"] = phase

        log_factor = 0
        return sample, log_factor


class PhaseReversalProposal(BaseGravitationalWaveTransientProposal):
    def __init__(self, priors, weight=1, fuzz=True, fuzz_sigma=1e-1):
        super(PhaseReversalProposal, self).__init__(priors, weight)
        self.fuzz = fuzz
        self.fuzz_sigma = fuzz_sigma
        if self.phase_key is None:
            raise SamplerError(
                f"{type(self).__name__} initialised without a phase prior"
            )

    def propose(self, chain):
        sample = chain.current_sample
        phase = sample[self.phase_key]
        sample[self.phase_key] = np.mod(phase + np.pi + self.epsilon, 2 * np.pi)
        log_factor = 0
        return sample, log_factor

    @property
    def epsilon(self):
        if self.fuzz:
            return np.random.normal(0, self.fuzz_sigma)
        else:
            return 0


class PolarisationReversalProposal(PhaseReversalProposal):
    def __init__(self, priors, weight=1, fuzz=True, fuzz_sigma=1e-3):
        super(PolarisationReversalProposal, self).__init__(
            priors, weight, fuzz, fuzz_sigma
        )
        self.fuzz = fuzz

    def propose(self, chain):
        sample = chain.current_sample
        psi = sample["psi"]
        sample["psi"] = np.mod(psi + np.pi / 2 + self.epsilon, np.pi)
        log_factor = 0
        return sample, log_factor


class PhasePolarisationReversalProposal(PhaseReversalProposal):
    def __init__(self, priors, weight=1, fuzz=True, fuzz_sigma=1e-1):
        super(PhasePolarisationReversalProposal, self).__init__(
            priors, weight, fuzz, fuzz_sigma
        )
        self.fuzz = fuzz

    def propose(self, chain):
        sample = chain.current_sample
        sample[self.phase_key] = np.mod(
            sample[self.phase_key] + np.pi + self.epsilon, 2 * np.pi
        )
        sample["psi"] = np.mod(sample["psi"] + np.pi / 2 + self.epsilon, np.pi)
        log_factor = 0
        return sample, log_factor


class StretchProposal(BaseProposal):
    """The Goodman & Weare (2010) Stretch proposal for an MCMC chain

    Implementation of the Stretch proposal using a sample drawn from the chain.
    We assume the form of g(z) from Equation (9) of [1].

    References
    ----------
    [1] Goodman & Weare (2010)
        https://ui.adsabs.harvard.edu/abs/2010CAMCS...5...65G/abstract

    """

    def __init__(self, priors, weight=1, subset=None, scale=2):
        super(StretchProposal, self).__init__(priors, weight, subset)
        self.scale = scale

    def propose(self, chain):
        sample = chain.current_sample

        # Draw a random sample
        rand = chain.random_sample

        return _stretch_move(sample, rand, self.scale, self.ndim, self.parameters)


def _stretch_move(sample, complement, scale, ndim, parameters):
    # Draw z
    u = np.random.rand()
    z = (u * (scale - 1) + 1) ** 2 / scale

    log_factor = (ndim - 1) * np.log(z)

    for key in parameters:
        sample[key] = complement[key] + (sample[key] - complement[key]) * z

    return sample, log_factor


class EnsembleProposal(BaseProposal):
    """Base EnsembleProposal class for ensemble-based swap proposals"""

    def __init__(self, priors, weight=1):
        super(EnsembleProposal, self).__init__(priors, weight)

    def __call__(self, chain, chain_complement):
        sample, log_factor = self.propose(chain, chain_complement)
        sample = self.apply_boundaries(sample)
        return sample, log_factor


class EnsembleStretch(EnsembleProposal):
    """The Goodman & Weare (2010) Stretch proposal for an Ensemble

    Implementation of the Stretch proposal using a sample drawn from complement.
    We assume the form of g(z) from Equation (9) of [1].

    References
    ----------
    [1] Goodman & Weare (2010)
        https://ui.adsabs.harvard.edu/abs/2010CAMCS...5...65G/abstract

    """

    def __init__(self, priors, weight=1, scale=2):
        super(EnsembleStretch, self).__init__(priors, weight)
        self.scale = scale

    def propose(self, chain, chain_complement):
        sample = chain.current_sample
        completement = chain_complement[
            np.random.randint(len(chain_complement))
        ].current_sample
        return _stretch_move(
            sample, completement, self.scale, self.ndim, self.parameters
        )


def get_default_ensemble_proposal_cycle(priors):
    return ProposalCycle([EnsembleStretch(priors)])


def get_proposal_cycle(string, priors, L1steps=1, warn=True):
    big_weight = 10
    small_weight = 5
    tiny_weight = 0.1

    if "gwA" in string:
        # Parameters for learning proposals
        learning_kwargs = dict(
            first_fit=1000, nsamples_for_density=10000, fit_multiplier=2
        )

        all_but_cal = [key for key in priors if "recalib" not in key]
        plist = [
            AdaptiveGaussianProposal(priors, weight=small_weight, subset=all_but_cal),
            DifferentialEvolutionProposal(
                priors, weight=small_weight, subset=all_but_cal
            ),
        ]

        if GMMProposal.check_dependencies(warn=warn) is False:
            raise SamplerError(
                "the gwA proposal_cycle required the GMMProposal dependencies"
            )

        if priors.intrinsic:
            intrinsic = PARAMETER_SETS["intrinsic"]
            plist += [
                AdaptiveGaussianProposal(priors, weight=big_weight, subset=intrinsic),
                DifferentialEvolutionProposal(
                    priors, weight=big_weight, subset=intrinsic
                ),
                KDEProposal(
                    priors, weight=big_weight, subset=intrinsic, **learning_kwargs
                ),
                GMMProposal(
                    priors, weight=big_weight, subset=intrinsic, **learning_kwargs
                ),
            ]

        if priors.extrinsic:
            extrinsic = PARAMETER_SETS["extrinsic"]
            plist += [
                AdaptiveGaussianProposal(priors, weight=small_weight, subset=extrinsic),
                DifferentialEvolutionProposal(
                    priors, weight=big_weight, subset=extrinsic
                ),
                KDEProposal(
                    priors, weight=big_weight, subset=extrinsic, **learning_kwargs
                ),
                GMMProposal(
                    priors, weight=big_weight, subset=extrinsic, **learning_kwargs
                ),
            ]

        if priors.mass:
            mass = PARAMETER_SETS["mass"]
            plist += [
                DifferentialEvolutionProposal(priors, weight=small_weight, subset=mass),
                GMMProposal(
                    priors, weight=small_weight, subset=mass, **learning_kwargs
                ),
            ]

        if priors.spin:
            spin = PARAMETER_SETS["spin"]
            plist += [
                DifferentialEvolutionProposal(priors, weight=small_weight, subset=spin),
                GMMProposal(
                    priors, weight=small_weight, subset=spin, **learning_kwargs
                ),
            ]
        if priors.precession:
            measured_spin = ["chi_1", "chi_2", "a_1", "a_2", "chi_1_in_plane"]
            plist += [
                AdaptiveGaussianProposal(
                    priors, weight=small_weight, subset=measured_spin
                ),
            ]

        if priors.mass and priors.spin:
            primary_spin_and_q = PARAMETER_SETS["primary_spin_and_q"]
            plist += [
                DifferentialEvolutionProposal(
                    priors, weight=small_weight, subset=primary_spin_and_q
                ),
            ]

        if getattr(priors, "tidal", False):
            tidal = PARAMETER_SETS["tidal"]
            plist += [
                DifferentialEvolutionProposal(
                    priors, weight=small_weight, subset=tidal
                ),
                PriorProposal(priors, weight=small_weight, subset=tidal),
            ]
        if priors.phase:
            plist += [
                PhaseReversalProposal(priors, weight=tiny_weight),
            ]
        if priors.phase and "psi" in priors.non_fixed_keys:
            plist += [
                CorrelatedPolarisationPhaseJump(priors, weight=tiny_weight),
                PhasePolarisationReversalProposal(priors, weight=tiny_weight),
            ]
        for key in ["time_jitter", "psi", "phi_12", "tilt_2", "lambda_1", "lambda_2"]:
            if key in priors.non_fixed_keys:
                plist.append(PriorProposal(priors, subset=[key], weight=tiny_weight))
        if "chi_1_in_plane" in priors and "chi_2_in_plane" in priors:
            in_plane = ["chi_1_in_plane", "chi_2_in_plane", "phi_12"]
            plist.append(UniformProposal(priors, subset=in_plane, weight=tiny_weight))
        if any("recalib_" in key for key in priors):
            calibration = [key for key in priors if "recalib_" in key]
            plist.append(PriorProposal(priors, subset=calibration, weight=small_weight))
    else:
        plist = [
            AdaptiveGaussianProposal(priors, weight=big_weight),
            DifferentialEvolutionProposal(priors, weight=big_weight),
            UniformProposal(priors, weight=tiny_weight),
            KDEProposal(priors, weight=big_weight, scale_fits=L1steps),
        ]
        if GMMProposal.check_dependencies(warn=warn):
            plist.append(GMMProposal(priors, weight=big_weight, scale_fits=L1steps))
        if NormalizingFlowProposal.check_dependencies(warn=warn):
            plist.append(
                NormalizingFlowProposal(priors, weight=big_weight, scale_fits=L1steps)
            )

    plist = remove_proposals_using_string(plist, string)
    return ProposalCycle(plist)


def remove_proposals_using_string(plist, string):
    mapping = dict(
        DE=DifferentialEvolutionProposal,
        AG=AdaptiveGaussianProposal,
        ST=StretchProposal,
        FG=FixedGaussianProposal,
        NF=NormalizingFlowProposal,
        KD=KDEProposal,
        GM=GMMProposal,
        PR=PriorProposal,
        UN=UniformProposal,
    )

    for element in string.split("no")[1:]:
        if element in mapping:
            plist = [p for p in plist if isinstance(p, mapping[element]) is False]
    return plist



import os
import shutil
import unittest

import bilby
from bilby.bilby_mcmc.sampler import Bilby_MCMC, BilbyMCMCSampler
from bilby.bilby_mcmc.utils import ConvergenceInputs
from bilby.core.sampler.base_sampler import SamplerError
import numpy as np
import pandas as pd


class TestBilbyMCMCSampler(unittest.TestCase):
    def setUp(self):
        default_kwargs = Bilby_MCMC.default_kwargs
        default_kwargs["target_nsamples"] = 100
        default_kwargs["L1steps"] = 1
        self.convergence_inputs = ConvergenceInputs(
            **{key: default_kwargs[key] for key in ConvergenceInputs._fields}
        )

        self.outdir = "bilby_mcmc_sampler_test"
        if os.path.isdir(self.outdir) is False:
            os.mkdir(self.outdir)

        def model(time, m, c):
            return time * m + c
        injection_parameters = dict(m=0.5, c=0.2)
        sampling_frequency = 10
        time_duration = 10
        time = np.arange(0, time_duration, 1 / sampling_frequency)
        N = len(time)
        sigma = np.random.normal(1, 0.01, N)
        data = model(time, **injection_parameters) + np.random.normal(0, sigma, N)
        likelihood = bilby.likelihood.GaussianLikelihood(time, data, model, sigma)

        # From hereon, the syntax is exactly equivalent to other bilby examples
        # We make a prior
        priors = dict()
        priors['m'] = bilby.core.prior.Uniform(0, 5, 'm')
        priors['c'] = bilby.core.prior.Uniform(-2, 2, 'c')
        priors = bilby.core.prior.PriorDict(priors)

        search_parameter_keys = ['m', 'c']
        use_ratio = False

        bilby.core.sampler.base_sampler._initialize_global_variables(
            likelihood,
            priors,
            search_parameter_keys,
            use_ratio,
        )

    def tearDown(self):
        if os.path.isdir(self.outdir):
            shutil.rmtree(self.outdir)

    def test_None_proposal_cycle(self):
        with self.assertRaises(SamplerError):
            BilbyMCMCSampler(
                convergence_inputs=self.convergence_inputs,
                proposal_cycle=None,
                beta=1,
                Tindex=0,
                Eindex=0,
                use_ratio=False
            )

    def test_default_proposal_cycle(self):
        sampler = BilbyMCMCSampler(
            convergence_inputs=self.convergence_inputs,
            proposal_cycle="default_noNFnoGMnoKD",
            beta=1,
            Tindex=0,
            Eindex=0,
            use_ratio=False
        )

        nsteps = 0
        while sampler.nsamples < 500:
            sampler.step()
            nsteps += 1
        self.assertEqual(sampler.chain.position, nsteps)
        self.assertEqual(sampler.accepted + sampler.rejected, nsteps)
        self.assertTrue(isinstance(sampler.samples, pd.DataFrame))


if __name__ == "__main__":
    unittest.main()
